def patch_singledispatch():
    "Patch functools.singledispatch to support Union types. Invoked only for python<=3.10"
    import sys
    if sys.version_info >= (3, 11):
        return
    import functools
    from functools import singledispatch as _singledispatch
    from types import UnionType

    def singledispatch(func):
        func = _singledispatch(func)
        _register = func.register

        def register(cls, func=None):
            from typing import get_type_hints
            if isinstance(cls, type):
                return _register(cls, func)
            if isinstance(cls, UnionType):
                def register_seq(func):
                    for t in cls.__args__:
                        _register(t, func)
                    return func
                return register_seq

            func = cls
            argname, cls = next(iter(get_type_hints(func).items()))
            if isinstance(cls, type):
                return _register(cls, func)
            if isinstance(cls, UnionType):
                for t in cls.__args__:
                    func = _register(t, func)
                return func

        func.register = register  # type: ignore
        return func
    setattr(functools, 'singledispatch', singledispatch)


patch_singledispatch()



import numpy as np
import tensorflow as tf
from keras.src.utils import tf_utils
from tensorflow import keras

from ..utils import warn
from .base import HLayerBase


class HBatchNormBase(HLayerBase):
    def __init__(
        self,
        axis=-1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):

        super().__init__(**kwargs)
        self.axis = [axis] if not isinstance(axis, (list, tuple)) else list(axis)
        self.momentum = tf.constant(momentum, dtype=self.dtype)
        self.epsilon = tf.constant(epsilon, dtype=self.dtype)
        self.center = center
        self.scale = scale
        self.beta_initializer = tf.keras.initializers.get(beta_initializer)
        self.gamma_initializer = tf.keras.initializers.get(gamma_initializer)
        self.moving_mean_initializer = tf.keras.initializers.get(
            moving_mean_initializer
        )
        self.moving_variance_initializer = tf.keras.initializers.get(
            moving_variance_initializer
        )
        self.beta_regularizer = tf.keras.regularizers.get(beta_regularizer)
        self.gamma_regularizer = tf.keras.regularizers.get(gamma_regularizer)
        self.beta_constraint = tf.keras.constraints.get(beta_constraint)
        self.gamma_constraint = tf.keras.constraints.get(gamma_constraint)
        self.norm_shape = None

    def post_build(self, input_shape):
        super().post_build(input_shape)
        self.axis: list = tf_utils.validate_axis(self.axis, input_shape)
        self._post_build(input_shape)  # type: ignore

    def _post_build(self, input_shape):
        self._reduction_axis = tuple([i for i in range(len(input_shape)) if i not in self.axis])
        output_shape = self.compute_output_shape(input_shape)
        shape = self.norm_shape or tuple([output_shape[i] for i in self.axis])

        if self.center and not getattr(self, "use_bias", False):
            warn(f'`center` in fused BatchNorm can only be used if `use_bias` is True. Setting center to False.', stacklevel=3)
            self.center = False

        if self.center:
            self.bn_beta: tf.Variable = self.add_weight(
                name="bn_beta",
                shape=shape,
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
            )
            self.moving_mean: tf.Variable = self.add_weight(
                name="moving_mean",
                shape=shape,
                initializer=self.moving_mean_initializer,
                trainable=False,
            )
        else:
            self.bn_beta = self.moving_mean = tf.Variable(0.0, trainable=False)
        if self.scale:
            self.bn_gamma: tf.Variable = self.add_weight(
                name="bn_gamma",
                shape=shape,
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
            )
            self.moving_variance: tf.Variable = self.add_weight(
                name="moving_variance",
                shape=shape,
                initializer=self.moving_variance_initializer,
                trainable=False,
            )
        else:
            self.bn_gamma = self.moving_variance = tf.Variable(1.0, trainable=False)

    @property
    @tf.function(jit_compile=True)
    def fused_kernel(self):  # type: ignore
        if not self.scale:
            return self.kernel
        scale = self.bn_gamma * tf.math.rsqrt(self.moving_variance + self.epsilon)
        return self.kernel * scale

    @property
    @tf.function(jit_compile=True)
    def fused_bias(self):  # type: ignore
        if not self.center:
            return self.bias
        scale = self.bn_gamma * tf.math.rsqrt(self.moving_variance + self.epsilon)
        return self.bias - self.moving_mean * scale + self.bn_beta

    def adapt_fused_bn_kernel_bw_bits(self, x: tf.Tensor):
        """Adapt the bitwidth of the kernel quantizer to the input tensor, such that each input is represented with approximately the same number of bits after fused batchnormalization."""
        if not self.scale:
            self.kq.adapt_bw_bits(self.kernel)
            return

        fbw = tf.identity(self.kq.fbw)
        self.kq.fbw.assign(tf.ones_like(fbw) * 32)
        z = self.forward(x, training=False, record_minmax=False)
        self.kq.fbw.assign(fbw)

        var = tf.math.reduce_variance(z, axis=self._reduction_axis)
        scale = self.bn_gamma * tf.math.rsqrt(var + self.epsilon)
        fused_kernel = self.kernel * scale
        self.kq.adapt_bw_bits(fused_kernel)


class FakeObj:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class HBatchNormalization(HBatchNormBase):

    def __init__(
        self,
        axis=-1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(
            axis=axis,
            momentum=momentum,
            epsilon=epsilon,
            center=center,
            scale=scale,
            beta_initializer=beta_initializer,
            gamma_initializer=gamma_initializer,
            moving_mean_initializer=moving_mean_initializer,
            moving_variance_initializer=moving_variance_initializer,
            beta_regularizer=beta_regularizer,
            gamma_regularizer=gamma_regularizer,
            beta_constraint=beta_constraint,
            gamma_constraint=gamma_constraint,
            **kwargs,
        )
        self._delayed_kernel_bits_adaption = True

    def post_build(self, input_shape):
        self.step_counter = tf.Variable(1, trainable=False, dtype=tf.int32, name="step_counter")
        self.use_bias = True
        axis: list = tf_utils.validate_axis(self.axis, input_shape)
        reduction_axis = tuple([i for i in range(len(input_shape)) if i not in axis])
        ker_shape = tuple(1 if i in reduction_axis else n_inp for i, n_inp in enumerate(input_shape))
        self._reduction_axis = reduction_axis
        self.norm_shape = ker_shape
        self.kernel = FakeObj(shape=ker_shape)
        r = super().post_build(input_shape)
        delattr(self, "use_bias")
        delattr(self, "kernel")
        return r

    def forward(self, x, training=None, record_minmax=False):

        if training:
            self.step_counter.assign_add(1)
            var = tf.math.reduce_variance(x, axis=self._reduction_axis, keepdims=True)
            mean = tf.math.reduce_mean(x, axis=self._reduction_axis, keepdims=True)
            self.moving_mean.assign_sub((self.moving_mean - mean) * (1 - self.momentum))
            self.moving_variance.assign_sub((self.moving_variance - var) * (1 - self.momentum))
        else:
            # correction = 1 - tf.pow(self.momentum, tf.cast(self.step_counter, self.dtype))
            mean = self.moving_mean
            var = self.moving_variance

        ker = self.bn_gamma * tf.math.rsqrt(var + self.epsilon)
        bias = self.bn_beta - mean / ker

        if self._do_adapt_kernel_bits and self._delayed_kernel_bits_adaption:
            self.kq.adapt_bw_bits(ker)
            self._delayed_kernel_bits_adaption = False

        qker = self.kq(ker, training=training)  # type: ignore
        qbias = self.paq(bias, training=training)  # type: ignore
        z = qker * x + qbias

        input_bw = self.input_bw
        if input_bw is not None:
            kernel_bw = self._kernel_bw(qker)
            bops = tf.reduce_sum(input_bw * kernel_bw)
            self.bops.assign(bops)
            bops_loss = tf.cast(bops, tf.float32) * self.beta
            self.add_loss(bops_loss)

        return self.paq(z, training=training, record_minmax=record_minmax)  # type: ignore

    def compute_output_shape(self, input_shape):
        return input_shape

    @property
    def compute_exact_bops(self):
        mean = self.moving_mean
        var = self.moving_variance
        ker = self.bn_gamma * tf.math.rsqrt(var + self.epsilon)
        qker = self.kq(ker, training=False)  # type: ignore
        kn, int_bits, fb = self.kq.get_bits_exact(qker)
        kernel_bw = int_bits + fb  # sign not considered for kernel
        input_bw = self.input_bw_exact
        bops = int(np.sum(input_bw * kernel_bw))
        self.bops.assign(tf.constant(bops, dtype=tf.float32))
        return bops



import os
import random

import numpy as np
import pytest
import tensorflow as tf
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras

from HGQ import get_default_paq_conf, set_default_paq_conf
from HGQ.layers import HActivation, HAdd, HConv1D, HConv2D, HDense, HQuantize, PConcatenate, PDropout, PFlatten, PMaxPool1D, PMaxPool2D, PReshape


def create_model(rnd_strategy: str, io_type: str):

    pa_config = get_default_paq_conf()
    pa_config['skip_dims'] = 'all' if io_type == 'io_stream' else 'batch'
    pa_config['rnd_strategy'] = rnd_strategy
    set_default_paq_conf(pa_config)

    inp = keras.Input(shape=(10, 10))
    qinp = HQuantize()(inp)
    x = PReshape((10, 10, 1))(qinp)  # 10x10
    x = HConv2D(5, (3, 3), activation='relu', padding='valid', parallel_factor=64)(x)  # 5x8x8
    x = PMaxPool2D((1, 2))(x)  # 5x8x4
    x = HActivation('relu')(x)
    mp = PMaxPool2D((4, 4))(x)  # 5x2x1
    x = PReshape((10,))(mp)  # 5
    x = HDense(10)(x)
    y = PReshape((20, 5))(qinp)  # 20x5
    y = HConv1D(5, 5, strides=5, padding='valid', parallel_factor=4)(y)  # 4x5
    po = PMaxPool1D(2)(y)  # 2x5
    ay = HActivation('relu')(po)
    y = PFlatten()(ay)  # 5
    y = PDropout(0.5)(y)
    y = HDense(10)(y)
    z = PFlatten()(qinp)  # 5
    x = HAdd()([x, y])  # 5
    xy = PConcatenate()([x, y])  # 5
    z = HDense(10, activation='relu')(z)  # 5
    z = HDense(10, activation=None)(z)  # 5
    out = PConcatenate()([xy, z])  # 5
    out = HActivation('relu')(out)
    horrible_model = keras.Model(inp, out)

    for layer in horrible_model.layers:
        # Randomize weight bitwidths
        if hasattr(layer, 'kq'):
            fbw: tf.Variable = layer.kq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 8, fbw.shape).astype(np.float32)))
        # And activation bitwidths
        if hasattr(layer, 'paq'):
            fbw: tf.Variable = layer.paq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 6, fbw.shape).astype(np.float32)))
        if hasattr(layer, 'bias') and layer.bias is not None:
            bias: tf.Variable = layer.bias
            bias.assign(tf.constant(np.random.uniform(-4, 4, layer.bias.shape).astype(np.float32)))
    return horrible_model


def get_data(N: int, sigma: float, max_scale: float, seed):
    rng = np.random.default_rng(seed)
    a1 = rng.normal(0, sigma, (N, 10, 10, 1)).astype(np.float32)
    a2 = rng.uniform(0, max_scale, (1, 10, 10, 1)).astype(np.float32)
    return (a1 * a2).astype(np.float32)


@pytest.mark.parametrize("N", [50000, 10])
@pytest.mark.parametrize("rnd_strategy", ['auto'])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("cover_factor", [0.25, 1.0])
@pytest.mark.parametrize("aggressive", [True, False])
@pytest.mark.parametrize("backend", ['vivado', 'vitis'])
@pytest.mark.parametrize("seed", [42])
def test_syn_large(N: int, rnd_strategy: str, io_type: str, cover_factor: float, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model(rnd_strategy=rnd_strategy, io_type=io_type)
    data = get_data(N, 8, 1, seed)

    run_model_test(model, cover_factor, data, io_type, backend, dir, aggressive)



import numpy as np
import tensorflow as tf
from keras.saving import register_keras_serializable
from keras.src.layers.pooling.base_pooling1d import Pooling1D
from keras.src.layers.pooling.base_pooling2d import Pooling2D
from keras.src.utils import conv_utils

from ..utils import apf_to_tuple, tuple_to_apf
from .base import ABSBaseLayer


@register_keras_serializable(package="HGQ")
class PLayerBase(ABSBaseLayer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_last_layer = False

    @property
    def act_bw(self):
        """Returns the bitwidth of the pre-activation values. Differentiable."""
        shape = (1,) + self.output_shape[1:]
        input_bw = tf.reshape(self.input_bw, shape)
        return tf.ensure_shape(input_bw, shape)

    @property
    def act_bw_exact(self) -> np.ndarray:
        """Returns the bitwidth of the pre-activation values. Differentiable."""
        shape = (1,) + self.output_shape[1:]
        return self.input_bw_exact.reshape(shape)


@register_keras_serializable(package="HGQ")
class Signature(PLayerBase):

    def __init__(self, keep_negative, bits, int_bits, **kwargs):
        super().__init__(**kwargs)
        if not hasattr(keep_negative, 'shape'):
            keep_negative = tf.constant(keep_negative, dtype=tf.int8)
        if not hasattr(bits, 'shape'):
            bits = tf.constant(bits, dtype=tf.int8)
        if not hasattr(int_bits, 'shape'):
            int_bits = tf.constant(int_bits, dtype=tf.int8)

        self.bits = bits
        self.int_bits = int_bits
        self.keep_negative = keep_negative

    def build(self, input_shape):
        self.built = True
        self.bits = tf.broadcast_to(self.bits, (1,) + input_shape[1:])
        self.int_bits = tf.broadcast_to(self.int_bits, (1,) + input_shape[1:])
        self.keep_negative = tf.broadcast_to(self.keep_negative, (1,) + input_shape[1:])

    @tf.function(jit_compile=True)
    def call(self, x, training=None, record_minmax=None):
        return x

    @property
    @tf.function(jit_compile=True)
    def act_bw(self):
        return tf.keras.backend.cast_to_floatx(self.bits)

    @property
    def act_bw_exact(self) -> np.ndarray:
        return self.bits.numpy()  # type: ignore

    @property
    def input_bw(self):
        raise ValueError('Signature layer does not have input_bw')

    def get_config(self):
        return {
            'name': self.name,
            'keep_negative': self.keep_negative.numpy().tolist(),  # type: ignore
            'bits': self.bits.numpy().tolist(),  # type: ignore
            'int_bits': self.int_bits.numpy().tolist(),  # type: ignore
        }


@register_keras_serializable(package="HGQ")
class PReshape(tf.keras.layers.Reshape, PLayerBase):
    pass


@register_keras_serializable(package="HGQ")
class PFlatten(tf.keras.layers.Flatten, PLayerBase):
    pass


@register_keras_serializable(package="HGQ")
class PConcatenate(tf.keras.layers.Concatenate, PLayerBase):

    @property
    @tf.function(jit_compile=True)
    def act_bw(self):
        """Returns the bitwidth of the pre-activation values. Differentiable."""
        shape = (1,) + self.output_shape[1:]
        return tf.concat(self.input_bw, axis=self.axis)

    @property
    def input_bw(self):
        assert len(self._inbound_nodes) <= 1, f"Layer {self.name} is reused {len(self._inbound_nodes)} times. This is not allowed."
        assert len(self._inbound_nodes[0].inbound_layers) > 1, "Concatenate layer should have at least two inputs"
        input_bws = [l.act_bw for l in self._inbound_nodes[0].inbound_layers]
        return tf.concat(input_bws, axis=self.axis)

    @property
    def act_bw_exact(self) -> np.ndarray:
        return np.concatenate([l.act_bw_exact for l in self._inbound_nodes[0].inbound_layers], axis=self.axis)


@register_keras_serializable(package="HGQ")
class PPool2D(PLayerBase, Pooling2D):

    def build(self, input_shape):
        super().build(input_shape)
        self._tf_data_format = conv_utils.convert_data_format(self.data_format, 4)

    @property
    def act_bw(self):
        act_bw = super().input_bw
        return tf.nn.max_pool2d(act_bw, ksize=self.pool_size, strides=self.strides, padding=self.padding.upper(), data_format=self._tf_data_format)

    @property
    def act_bw_exact(self) -> np.ndarray:
        act_bw = super().input_bw_exact
        return tf.nn.max_pool2d(act_bw, ksize=self.pool_size, strides=self.strides, padding=self.padding.upper(), data_format=self._tf_data_format).numpy()


@register_keras_serializable(package="HGQ")
class PPool1D(PLayerBase, Pooling1D):

    def build(self, input_shape):
        super().build(input_shape)
        self._tf_data_format = conv_utils.convert_data_format(self.data_format, 3)

    @property
    def act_bw(self):
        act_bw = super().input_bw
        return tf.nn.max_pool1d(act_bw, ksize=self.pool_size, strides=self.strides, padding=self.padding.upper(), data_format=self._tf_data_format)

    @property
    def act_bw_exact(self) -> np.ndarray:
        act_bw = super().input_bw_exact
        return tf.nn.max_pool1d(act_bw, ksize=self.pool_size, strides=self.strides, padding=self.padding.upper(), data_format=self._tf_data_format).numpy()


@register_keras_serializable(package="HGQ")
class PMaxPooling2D(PPool2D, tf.keras.layers.MaxPool2D):
    pass


@register_keras_serializable(package="HGQ")
class PAveragePooling2D(PPool2D, tf.keras.layers.AvgPool2D):
    pass


@register_keras_serializable(package="HGQ")
class PMaxPooling1D(PPool1D, tf.keras.layers.MaxPool1D):
    pass


@register_keras_serializable(package="HGQ")
class PAveragePooling1D(PPool1D, tf.keras.layers.AvgPool1D):
    pass


PMaxPool2D = PMaxPooling2D
PAvgPool2D = PAveragePooling2D
PMaxPool1D = PMaxPooling1D
PAvgPool1D = PAveragePooling1D


@register_keras_serializable(package="HGQ")
class PDropout(PLayerBase, tf.keras.layers.Dropout):
    pass


@register_keras_serializable(package="HGQ")
class PPermute(PLayerBase, tf.keras.layers.Permute):
    pass



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

import HGQ

sys.path.insert(0, os.path.abspath('../'))

project = 'High Granularity Quantization'
copyright = '2023, Chang Sun'
author = 'Chang Sun'
release = str(HGQ.__version__)
version = release

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

myst_enable_extensions = [
    "amsmath",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]


autosummary_generate = True

extensions = ['myst_parser', 'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.napoleon', 'sphinx_rtd_theme']

source_suffix = ['.rst', '.md']

templates_path = ['_templates']
exclude_patterns = ['_build']

html_logo = "_static/logo.svg"
html_theme_options = {
    'logo_only': True,
    'display_version': True,
}

html_context = {
    'display_github': True,  # Integrate GitHub
    'github_user': 'calad0i',  # Username
    'github_repo': "HGQ",  # Repo name
    'github_version': 'master',  # Version
    'conf_py_path': '/docs/',  # Path in the checkout to the docs root
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']
html_favicon = '_static/icon.svg'

html_css_files = [
    'custom.css',
]



import warnings

import numpy as np
import pytest
import tensorflow as tf
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras

from HGQ import get_default_paq_conf, set_default_paq_conf
from HGQ.layers import HDense, Signature


def create_model():
    # Here is an extremely terrible example of how to use Signature layer.
    # This is for test purpose only. Do NOT use it as a reference.

    dense = keras.layers.Dense(8)
    inp = keras.Input(shape=(8,))
    out = Signature(0, 4, 4)(inp)
    out = dense(out)
    model = keras.Model(inp, out)
    dense.kernel.assign(tf.constant(np.eye(8).astype(np.float32)))
    return model


def get_data(N: int, seed):
    rng = np.random.default_rng(seed)
    data = rng.integers(0, 2**4, (N, 8)).astype(np.float32)
    return data


@pytest.mark.parametrize("N", [1000])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("aggressive", [True, False])
@pytest.mark.parametrize("backend", ['vivado'])
@pytest.mark.parametrize("seed", [42])
def test_syn_hlayers(N: int, io_type: str, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model()
    data = get_data(N, seed)

    # with pytest.warns(UserWarning, match=r'Layer batch_normalization(?:|_\d+) is unknown. Assuming infinite produced bitwidth') as record:
    run_model_test(model,
                   None,
                   data,
                   io_type,
                   backend,
                   dir,
                   aggressive,
                   test_gard=False,
                   )
    # assert record



import os
import random

import numpy as np
import pytest
import tensorflow as tf
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras

from HGQ import get_default_paq_conf, set_default_paq_conf
from HGQ.layers import HActivation, HAdd, HConv1D, HConv2D, HDense, HQuantize, PConcatenate, PDropout, PFlatten, PMaxPool1D, PMaxPool2D, PReshape


def create_model(rnd_strategy: str, io_type: str):

    pa_config = get_default_paq_conf()
    pa_config['skip_dims'] = 'all' if io_type == 'io_stream' else 'batch'
    pa_config['rnd_strategy'] = rnd_strategy
    set_default_paq_conf(pa_config)

    model = keras.Sequential([
        HQuantize(input_shape=(6, 6, 1), name='inp_q'),
        HConv2D(4, 2, activation='relu'),
        PMaxPool2D(2),
        PReshape((8, 2)),
        HConv1D(4, 3, activation='relu', padding='same'),
        PMaxPool1D(2),
        PFlatten(),
        HDense(10),
        HActivation('sigmoid')
    ])

    for layer in model.layers:
        # Randomize weight bitwidths
        if hasattr(layer, 'kq'):
            fbw: tf.Variable = layer.kq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 8, fbw.shape).astype(np.float32)))
        # And activation bitwidths
        if hasattr(layer, 'paq'):
            fbw: tf.Variable = layer.paq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 6, fbw.shape).astype(np.float32)))
        if hasattr(layer, 'bias') and layer.bias is not None:
            bias: tf.Variable = layer.bias
            bias.assign(tf.constant(np.random.uniform(-4, 4, layer.bias.shape).astype(np.float32)))
    return model


def get_data(N: int, sigma: float, max_scale: float, seed):
    rng = np.random.default_rng(seed)
    a1 = rng.normal(0, sigma, (N, 6, 6, 1)).astype(np.float32)
    a2 = rng.uniform(0, max_scale, (1, 6, 6, 1)).astype(np.float32)
    return (a1 * a2).astype(np.float32)


@pytest.mark.parametrize("N", [50000, 10])
@pytest.mark.parametrize("rnd_strategy", ['auto', 'standard_round', 'floor'])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("cover_factor", [0.25, 1.0])
@pytest.mark.parametrize("aggressive", [True, False])
@pytest.mark.parametrize("backend", ['vivado', 'vitis'])
@pytest.mark.parametrize("seed", [114514, 42, 1919810])
def test_syn_small(N: int, rnd_strategy: str, io_type: str, cover_factor: float, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model(rnd_strategy=rnd_strategy, io_type=io_type)
    data = get_data(N, 8, 1, seed)

    run_model_test(model, cover_factor, data, io_type, backend, dir, aggressive)


if __name__ == '__main__':
    test_syn_small(10, 'auto', 'io_parallel', 0.49, True, 'vivado', 1919810)



import json
import os
import random
import shutil
from warnings import warn

import numpy as np
import pytest
import tensorflow as tf

tf.config.experimental_run_functions_eagerly(True)  # noqa
tf.config.threading.set_inter_op_parallelism_threads(1)  # noqa
tf.config.threading.set_intra_op_parallelism_threads(1)  # noqa

from hls4ml.converters import convert_from_keras_model
from tensorflow import keras

from HGQ import trace_minmax
from HGQ.proxy import FixedPointQuantizer, to_proxy_model

tf.get_logger().setLevel('ERROR')


def get_test_dir() -> str:
    cur_test = os.environ.get('PYTEST_CURRENT_TEST')
    if not cur_test:
        cur_test = 'test_run'
    cur_test = cur_test.replace('::', '_').replace('(call)', '')
    test_root = os.environ.get('TEST_ROOT_DIR')
    if not test_root:
        test_root = '/tmp/unit_test'
    test_dir = os.path.join(test_root, cur_test)
    os.makedirs(test_dir, exist_ok=True)
    return test_dir


def _run_synth_match_test(proxy: keras.Model, data, io_type: str, backend: str, dir: str, cond=None):

    output_dir = dir + '/hls4ml_prj'
    hls_model = convert_from_keras_model(
        proxy,
        io_type=io_type,
        output_dir=output_dir,
        backend=backend,
        hls_config={'Model': {'Precision': 'fixed<1,0>', 'ReuseFactor': 1}}
    )
    hls_model.compile()

    data_len = data.shape[0] if isinstance(data, np.ndarray) else data[0].shape[0]
    # Multiple output case. Check each output separately
    if len(proxy.outputs) > 1:  # type: ignore
        r_proxy: list[np.ndarray] = [x.numpy() for x in proxy(data)]  # type: ignore
        r_hls: list[np.ndarray] = hls_model.predict(data)  # type: ignore
        r_hls = [x.reshape(r_proxy[i].shape) for i, x in enumerate(r_hls)]
    else:
        r_proxy: list[np.ndarray] = [proxy(data).numpy()]  # type: ignore
        r_hls: list[np.ndarray] = [hls_model.predict(data).reshape(r_proxy[0].shape)]  # type: ignore

    errors = []
    for i, (p, h) in enumerate(zip(r_proxy, r_hls)):
        try:
            if cond is None:
                mismatch_ph = p != h
                assert np.sum(mismatch_ph) == 0, f"Proxy-HLS4ML mismatch for out {i}: {np.sum(np.any(mismatch_ph,axis=1))} out of {data_len} samples are different. Sample: {p[mismatch_ph].ravel()[:5]} vs {h[mismatch_ph].ravel()[:5]}"
            else:
                cond(p, h)
        except AssertionError as e:
            errors.append(e)
    if len(errors) > 0:
        msgs = [str(e) for e in errors]
        raise AssertionError('\n'.join(msgs))


def _run_model_sl_test(model: keras.Model, proxy: keras.Model, data, output_dir: str):
    model.save(output_dir + '/keras.h5')
    model_loaded: keras.Model = keras.models.load_model(output_dir + '/keras.h5')  # type: ignore

    proxy.save(output_dir + '/proxy.h5')
    proxy_loaded: keras.Model = keras.models.load_model(output_dir + '/proxy.h5')  # type: ignore
    for l1, l2 in zip(proxy.layers, proxy_loaded.layers):
        if not isinstance(l1, FixedPointQuantizer):
            continue
        assert l1.overrides == l2.overrides, f"Overrides mismatch for layer {l1.name}"

    assert tf.reduce_all(model(data) == model_loaded(data)), f"Model premdiction mismatch"
    assert tf.reduce_all(proxy(data) == proxy_loaded(data)), f"Proxy prediction mismatch"


def _run_model_proxy_match_test(model: keras.Model, proxy: keras.Model, data, cover_factor: float | None):
    nof_outputs = len(model.outputs)  # type: ignore
    if nof_outputs > 1:
        r_keras: list[np.ndarray] = [x.numpy() for x in model(data)]  # type: ignore
        r_proxy: list[np.ndarray] = [x.numpy() for x in proxy(data)]  # type: ignore
    else:
        r_keras: list[np.ndarray] = [model(data).numpy()]  # type: ignore
        r_proxy: list[np.ndarray] = [proxy(data).numpy()]  # type: ignore

    errors = []
    for i, (k, p) in enumerate(zip(r_keras, r_proxy)):
        mismatch_kp = k != p
        try:
            assert np.sum(mismatch_kp) == 0, f"Keras-Proxy mismatch for out {i}: {np.sum(np.any(mismatch_kp,axis=1))} out of {data.shape[0]} samples are different. Sample: {k[mismatch_kp].ravel()[:5]} vs {p[mismatch_kp].ravel()[:5]}"
        except AssertionError as e:
            errors.append(e)

    if cover_factor is None or cover_factor >= 1.0:
        if not len(errors) == 0:
            raise AssertionError('\n'.join([str(e) for e in errors]))
    else:
        if len(errors) == 0:
            warn(f"Keras-Proxy perfect match when overflow should happen: cover_factor={cover_factor}.")


def _run_gradient_test(model, data):
    with tf.GradientTape() as tape:
        out = model(data, training=True)
        if isinstance(out, list):
            loss = tf.reduce_mean([tf.reduce_sum(tf.abs(x)) for x in out])
        else:
            loss = tf.reduce_sum(tf.abs(out))

    trainable_weights = model.trainable_weights
    grad = tape.gradient(loss, trainable_weights)

    for w, g in zip(trainable_weights, grad):
        assert tf.reduce_any(g != 0), f"Gradient for {w.name} is zero"


def run_model_test(model: keras.Model, cover_factor: float | None, data, io_type: str, backend: str, dir: str, aggressive: bool, cond=None, skip_sl_test=False, test_gard=False):
    data_len = data.shape[0] if isinstance(data, np.ndarray) else data[0].shape[0]
    if test_gard:
        _run_gradient_test(model, data)
    if cover_factor is not None:
        trace_minmax(model, data, cover_factor=cover_factor, bsz=data_len)
    proxy = to_proxy_model(model, aggressive=aggressive, unary_lut_max_table_size=4096)
    try:
        if not skip_sl_test:
            _run_model_sl_test(model, proxy, data, dir)
        _run_model_proxy_match_test(model, proxy, data, cover_factor)
        _run_synth_match_test(proxy, data, io_type, backend, dir, cond=cond)
    except AssertionError as e:
        raise e
    except Warning as w:
        warn(w)
    else:
        shutil.rmtree(dir)


def set_seed(seed: int):
    seed = seed
    os.environ['RANDOM_SEED'] = f'{seed}'
    np.random.seed(seed)
    tf.random.set_seed(seed)
    random.seed(seed)



import numpy as np
import tensorflow as tf
from tensorflow import keras

from ..layers import HLayerBase
from ..utils import warn


class FreeBOPs(tf.keras.callbacks.Callback):
    def __init__(self):
        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        assert self.model is not None
        logs['bops'] = self._bops(self.model)  # type: ignore

    def _bops(self, model):
        bops = 0
        for layer in model.layers:
            if hasattr(layer, 'bops'):
                bops += layer.bops.numpy()
                continue
            if isinstance(layer, keras.Model):
                bops += self._bops(layer)
        return bops


class ResetMinMax(tf.keras.callbacks.Callback):
    def on_epoch_end(self, epoch, logs=None):
        assert self.model is not None
        self.reset_minmax(self.model)

    @staticmethod
    def reset_minmax(model):
        for layer in model.layers:
            if isinstance(layer, keras.Model):
                ResetMinMax.reset_minmax(layer)
            if isinstance(layer, HLayerBase):
                layer.reset_minmax()


class CalibratedBOPs(tf.keras.callbacks.Callback):
    def __init__(self, calibration_data, bsz=None):
        self.calibration_data = calibration_data
        self.bsz = bsz

    def on_epoch_end(self, epoch, logs=None):
        assert isinstance(logs, dict)
        assert isinstance(self.model, keras.models.Model)

        data = self.calibration_data
        bsz = self.bsz or len(data)
        bops = trace_minmax(self.model, data, bsz=bsz, verbose=False)
        logs['bops'] = bops


def trace_minmax(model, dataset, bsz=16384, verbose=True, return_predictions=False, no_bops_computation=False, rst=True, cover_factor=1.0):
    if rst:
        for layer in model.layers:
            if isinstance(layer, HLayerBase):
                layer.reset_minmax()
                layer.record_minmax = True

    r = []
    if isinstance(dataset, list):
        length = len(dataset[0])
        for i in range(0, length, bsz):
            r.append(model([d[i:i + bsz] for d in dataset], training=False))
    else:
        for i in range(0, dataset.shape[0], bsz):
            r.append(model(dataset[i:i + bsz], training=False))

    if cover_factor != 1.0:
        assert cover_factor > 0.
        if cover_factor < 1:
            warn(f'cover_factor<1.0 will likely to result in overflows.')
        for layer in model.layers:
            if not isinstance(layer, HLayerBase):
                continue

            aq = layer.paq
            _min, _max = aq._min.numpy(), aq._max.numpy()  # type: ignore
            mean = (_min + _max) / 2
            diff = (_max - _min) / 2
            _min = mean - diff * cover_factor
            _max = mean + diff * cover_factor

            aq._min.assign(aq(tf.constant(_min)))  # type: ignore
            aq._max.assign(aq(tf.constant(_max)))  # type: ignore

    for layer in model.layers:
        if not hasattr(layer, 'activation'):
            continue
        if not isinstance(layer, HLayerBase):
            continue
        aq = layer.paq
        if layer.activation is tf.keras.activations.softmax:
            # softmax_lut behaves slightly differently from softmax, which is more likely to generate exact 1s. Overflows may occur with traced min-max.
            aq._min.assign(tf.zeros_like(aq._min))  # type: ignore
            aq._max.assign(tf.ones_like(aq._max))  # type: ignore

        if layer.activation is tf.keras.activations.sigmoid:
            # <0 or >1 for sigmoid does not make sense. Undo the effect of cover_factor.
            aq._min.assign(tf.maximum(aq._min, tf.zeros_like(aq._min)))  # type: ignore
            aq._max.assign(tf.minimum(aq._max, tf.ones_like(aq._max)))  # type: ignore

        if layer.activation is tf.keras.activations.tanh:
            # <-1 or >1 for tanh does not make sense. Undo the effect of cover_factor.
            aq._min.assign(tf.maximum(aq._min, -tf.ones_like(aq._min)))  # type: ignore
            aq._max.assign(tf.minimum(aq._max, tf.ones_like(aq._max)))  # type: ignore

    if no_bops_computation:
        return -1

    bops = 0
    for layer in model.layers:
        if isinstance(layer, HLayerBase):
            layer.record_minmax = False
            dbops = layer.compute_exact_bops  # type: ignore
            bops += dbops
            if verbose:
                print(f'{layer.name}: {dbops}')

    if return_predictions:
        return bops, np.concatenate(r, axis=0)

    return bops



from collections import abc

import tensorflow as tf
from keras import constraints, initializers, regularizers
from keras.layers import Dense, Dropout, Layer, MultiHeadAttention, Softmax
from keras.src.utils import tf_utils

from ..base import ABSBaseLayer, HLayerBase
from ..dense import HDense


class HMultiHeadAttention(Layer):
    def __init__(
        self,
        num_heads,
        key_dim,
        value_dim=None,
        dropout=0.0,
        use_bias=True,
        output_shape=None,
        attention_axes=None,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.supports_masking = True
        self._num_heads = num_heads
        self._key_dim = key_dim
        self._value_dim = value_dim if value_dim else key_dim
        self._dropout = dropout
        self._use_bias = use_bias
        self._output_shape = output_shape
        self._kernel_initializer = initializers.get(kernel_initializer)
        self._bias_initializer = initializers.get(bias_initializer)
        self._kernel_regularizer = regularizers.get(kernel_regularizer)
        self._bias_regularizer = regularizers.get(bias_regularizer)
        self._activity_regularizer = regularizers.get(activity_regularizer)
        self._kernel_constraint = constraints.get(kernel_constraint)
        self._bias_constraint = constraints.get(bias_constraint)
        if attention_axes is not None and not isinstance(
            attention_axes, abc.Sized
        ):
            self._attention_axes = (attention_axes,)
        else:
            self._attention_axes = attention_axes
        self._built_from_signature = False
        self._query_shape, self._key_shape, self._value_shape = None, None, None

    def attn_build(self, q_dim, k_dim, v_dim):
        with tf_utils.maybe_init_scope(self):  # type: ignore
            self.to_q = Dense(self._key_dim * self._num_heads, use_bias=self._use_bias, name=f"{self.name}_to_q")
            self.to_k = Dense(self._key_dim * self._num_heads, use_bias=self._use_bias, name=f"{self.name}_to_k")
            self.to_v = Dense(self._value_dim * self._num_heads, use_bias=self._use_bias, name=f"{self.name}_to_v")
            self.to_out = Dense(v_dim, use_bias=self._use_bias, name=f"{self.name}_to_out")
            self.dropout = Dropout(rate=self._dropout)
            self.norm = Softmax(axis=-1)
        self._query_shape = q_dim
        self._key_shape = k_dim
        self._value_shape = v_dim
        self._built_from_signature = True

    def compute_output_shape(self, input_shape):
        return tf.TensorShape((None, None, self._value_dim * self._num_heads))

    def call(self, q, k, v, attention_mask=None, training=None, return_attention_scores=False):
        if not self._built_from_signature:
            self.attn_build(q.shape[-1], k.shape[-1], v.shape[-1])

        # [B, N, HD]
        Q = self.to_q(q)
        K = self.to_k(k)
        V = self.to_v(v)

        # [B, N, H, D]
        Q = tf.reshape(Q, [*Q.shape[:2], self._num_heads, self._key_dim])
        K = tf.reshape(K, [*K.shape[:2], self._num_heads, self._key_dim])
        V = tf.reshape(V, [*V.shape[:2], self._num_heads, self._value_dim])

        # [B, H, N, D]
        Q = tf.transpose(Q, [0, 2, 1, 3])
        K = tf.transpose(K, [0, 2, 1, 3])
        V = tf.transpose(V, [0, 2, 1, 3])
        print(Q.shape, K.shape, V.shape)

        # [B, H, NQ, NK]
        QK = tf.matmul(Q, K, transpose_b=True)
        QK = QK / tf.math.sqrt(tf.cast(K.shape[-1], tf.float32))
        print(QK.shape)

        if attention_mask is not None:
            attention_mask = attention_mask[:, None]
            if attention_mask.dtype == tf.bool:
                attention_mask = tf.cast(~attention_mask, QK.dtype) * -1e9  # type: ignore
            QK = QK + attention_mask

        QK = self.norm(QK)

        # [B, H, NQ, DV]
        out = tf.matmul(QK, V)
        print(out.shape)

        # [B, N, H, D]
        out = tf.transpose(out, [0, 2, 1, 3])
        out = tf.reshape(out, [*out.shape[:2], self._value_dim * self._num_heads])
        out = self.to_out(out)
        if return_attention_scores:
            return out, QK
        return out



from collections.abc import Callable
from copy import copy, deepcopy
from functools import partial, singledispatch, singledispatchmethod
from typing import Any

import numpy as np
import tensorflow as tf
from keras.src.engine.keras_tensor import KerasTensor
from keras.src.engine.node import Node
from tensorflow import keras

from ..layers import HLayerBase, HQuantize, PDropout, PLayerBase, Signature
from ..layers.base import ABSBaseLayer
from ..utils import warn
from .fixed_point_quantizer import FixedPointQuantizer
from .precision_derivation import register_qconf
from .unary_lut import xfr_to_unary_lut


def get_all_nodes(model: keras.Model) -> set[Node]:
    """Get all nodes in the model as a set."""
    nodes = set()
    layers = set(model.layers)
    for layer in model.layers:
        for node in layer._inbound_nodes:
            if node.layer in layers:
                nodes.add(node)
        for node in layer._outbound_nodes:
            if node.layer in layers:
                nodes.add(node)
    return nodes


def solve_dependencies(model: keras.Model):
    """Given a keras model, return the input nodes, output nodes and a list of (layer, requires, provides) tuples. Requires is a list of nodes that are parents of the layer, provides is the node that is the output of the layer."""
    inp_tensors: list[KerasTensor] = model.inputs  # type:ignore
    out_tensors: list[KerasTensor] = model.outputs  # type:ignore

    input_nodes: list[Node] = [t.node for t in inp_tensors]  # type:ignore
    output_nodes: list[Node] = [t.node for t in out_tensors]  # type:ignore

    nodes = get_all_nodes(model)

    dependencies_dict: dict[Node, tuple[keras.layers.Layer, list[Node]]] = {}
    "List of (layer, requires, provides) tuples; requires is a list of nodes, provides is a single node."

    for node in nodes:
        if node.is_input:
            continue
        layer = node.layer  # layer that is called on the node
        requires = list(node.parent_nodes)
        provides = node
        dependencies_dict[provides] = (layer, requires)
    return input_nodes, output_nodes, dependencies_dict


def get_weight(layer: keras.layers.Layer, name: str):
    """Given a layer and a weight name, return the weight. The weight name may or may not contain the layer name. If the number index is missing, it is assumed to be 0."""
    if '/' in name:
        name = name.split('/')[-1]
    if ':' not in name:
        name = f'{name}:0'
    for w in layer.weights:
        if name in w.name:
            return w
    return None


def copy_fused_weights(src: keras.layers.Layer, dst: keras.layers.Layer):
    """For HGQ layers, some layers may have different fused weights for kernel and bias (Processed weights are deployment). This function copies the fused kernel and bias to the keras proxy."""
    if hasattr(dst, 'kernel') and dst.kernel is not None:
        if (k := getattr(src, 'fused_qkernel', None)) is not None:
            dst.kernel.assign(k)
        else:
            dst.kernel.assign(src.kernel)
    if hasattr(dst, 'bias') and dst.bias is not None:
        if (b := getattr(src, 'fused_qbias', None)) is not None:
            dst.bias.assign(b)
        else:
            dst.bias.assign(src.bias)
    for dst_w in dst.weights:
        if 'kernel:0' in dst_w.name or 'bias:0' in dst_w.name:
            continue
        src_w = get_weight(src, dst_w.name)
        if src_w is not None:
            dst_w.assign(src_w)
        else:
            warn(f'No weight found for layer {src.name} corresponding to {dst_w.name}.')


class Namer:
    """Helper class to generate unique names for layers, if one being used multiple times."""

    def __init__(self):
        self.used_names: set[str] = set()

    def next_name(self, name: str):
        i = 0
        if name in self.used_names:
            while f'{name}_{i}' in self.used_names:
                i += 1
            name = f'{name}_{i}'
        self.used_names.add(name)
        return name


def _get_out_tensor(
    dependencies_dict: dict[Node, tuple[keras.layers.Layer, list[Node]]],
    satisfied: dict[Node, Any],
    node: Node,
    namer: Namer | None = None,
    layer_xformer: Callable[[keras.layers.Layer], keras.layers.Layer] = lambda x: x,
):
    """Given a node, get the output tensor provided generated by the node."""

    assert node not in satisfied, f'Node {node} already satisfied.'

    namer = namer or Namer()
    layer, requires = dependencies_dict[node]
    inps = []
    for req in requires:
        if req in satisfied:
            inps.append(satisfied[req])
        else:
            inp = _get_out_tensor(dependencies_dict, satisfied, req, namer, layer_xformer)
            inps.append(inp)
    if len(inps) == 1:
        inps = inps[0]

    satisfied[node] = apply_layer(layer, inps, namer=namer, layer_xformer=layer_xformer)
    return satisfied[node]


@singledispatch
def apply_layer(
    layer: keras.layers.Layer | None,
    inp_tensors: list[tf.Tensor] | tf.Tensor,
    namer: None | Namer = None,
    # satisfied: dict[Node, Any] | None = None,
    layer_xformer: Callable[[keras.layers.Layer], keras.layers.Layer] = lambda x: x,
):
    """
    Apply a layer to the input tensors:
      - if the layer is a model, apply the model recursively in a flattened manner.
      - if the layer is a keras layer, transform it by layer_transformer, until the exactly same reference is returned.
    """

    layer_xf = layer_xformer(layer)
    n = 0
    namer = namer or Namer()
    while layer_xf is not layer:
        if isinstance(layer_xf, keras.Model):
            # layer_transformer may return a keras.Model from a layer.
            return apply_layer(layer_xf, inp_tensors, namer=namer, layer_xformer=layer_xformer)
        assert n < 1024, f'layer_transformer does not converge for layer (name: {layer.name if layer is not None else None}).'
        layer = layer_xf
        layer_xf = layer_xformer(layer)
        n += 1

    if layer_xf is not None:
        name = namer.next_name(layer_xf.name)
        layer_xf._name = name
        # Remove all inbound and outbound nodes to clean up the graph.
        layer_xf = copy(layer_xf)
        layer_xf._inbound_nodes = []
        layer_xf._outbound_nodes = []
        return layer_xf(inp_tensors)
    else:
        return inp_tensors


@apply_layer.register
def _(
    layer: keras.Model,
    inp_tensors: list[tf.Tensor] | tf.Tensor,
    namer: None | Namer = None,
    # satisfied: dict[Node, Any] | None = None,
    layer_xformer: Callable[[keras.layers.Layer], keras.layers.Layer] = lambda x: x,
):
    model = layer
    if not model.built:
        if isinstance(inp_tensors, (list, tuple)):
            model.build([t.shape for t in inp_tensors])
        else:
            model.build(inp_tensors.shape)

    namer = namer or Namer()
    inp_tensors = [inp_tensors] if not isinstance(inp_tensors, (list, tuple)) else inp_tensors
    input_nodes, output_nodes, dependencies_dict = solve_dependencies(model)
    satisfied = {node: tensor for node, tensor in zip(input_nodes, inp_tensors)}
    get_out_tensor = partial(_get_out_tensor, dependencies_dict, satisfied, namer=namer, layer_xformer=layer_xformer)
    outs = [get_out_tensor(output_node) for output_node in output_nodes]
    return outs[0] if len(outs) == 1 else outs


def convert_model(model: keras.Model, layer_xformer: Callable[[keras.layers.Layer], keras.layers.Layer] = lambda x: x):
    """For a keras model, convert each layer with layer_transformer, flatten all layers (remove all keras.Model as layers), and return a new model.

    Args:
        model (keras.Model): Input keras model.
        layer_transformer: A function that takes a keras layer and returns a keras layer (can be a keras.Model). For the final set of layer, it must return the same reference, rather than a copy, or the conversion will never end.

    Returns:
        keras.Model: A new keras model.

    Examples:
        Simply flatten a keras model: `convert_model(model)`
        Flatten a keras model, return a new model with all
    """

    inputs: list[tf.Tensor] = [keras.layers.Input(shape=t.shape[1:]) for t in model.inputs]  # type: ignore
    if len(inputs) == 1:
        inputs = inputs[0]
    outs = apply_layer(model, inputs, layer_xformer=layer_xformer)
    return keras.Model(inputs, outs)


@singledispatch
def to_keras_layer(layer):
    raise TypeError(f'No matching overload for layer type {type(layer)}. Signatures available: {to_keras_layer.registry.keys()}')


@to_keras_layer.register
def _(layer: ABSBaseLayer):
    """Given a HGQ layer, return the corresponding keras layer.

    Example:
        HDense -> Dense
        HQuantize -> Raise TypeError
    """

    assert not isinstance(layer, (HQuantize, Signature)), 'to_keras_layers should not be called on HQuantize or Signature layers.'

    if hasattr(layer, 'get_keras_config'):
        conf = layer.get_keras_config()
    else:
        conf = layer.get_config()

    if not isinstance(layer, keras.layers.Activation):
        if 'activation' in conf and conf['activation'] != 'linear':
            # Activation will be processed separately for non-activation layers.
            conf['activation'] = 'linear'
    else:
        # Prevent custom activation crashing conversion.
        conf['activation'] = layer.activation

    cls_name = layer.__class__.__name__[1:]
    if hasattr(keras.layers, cls_name):
        layer_cls = getattr(keras.layers, cls_name)
    elif hasattr(keras.layers, cls_name.replace('BatchNorm', '')):
        layer_cls = getattr(keras.layers, cls_name.replace('BatchNorm', ''))
    else:
        raise RuntimeError(f'Unknown layer type {layer.__class__.__name__}: no corresponding keras layer found.')
    klayer = layer_cls.from_config(conf)
    klayer.build(layer.input_shape)
    copy_fused_weights(layer, klayer)
    return klayer


class ProxyLayerXFormer:
    def __init__(self, SAT='WRAP', use_uniary_lut=False):
        self.SAT = SAT
        self.use_uniary_lut = use_uniary_lut

    @singledispatchmethod
    def get_kbiRS(self, layer: HLayerBase, **kwargs):
        q = layer.paq
        pos_only = kwargs.get('pos_only', False)
        k, i, f = q.get_bits_exact(pos_only=pos_only)
        k, b, i = k, k + i + f, k + i
        if q.rnd_strategy != 3 and not layer.can_bias_cover_rnd:
            RND = 'RND'
        else:
            RND = 'TRN'
        return k, b, i, RND, self.SAT

    @singledispatchmethod
    def __call__(self, layer: keras.layers.Layer | None) -> keras.layers.Layer | None:
        if layer is None:
            return None
        try:
            from hls4ml.converters.keras_to_hls import layer_handlers
        except ImportError:
            return layer
        cls_name = layer.__class__.__name__
        assert cls_name in layer_handlers, f'Layer {layer.name}: {cls_name} is not supported by hls4ml.'
        return layer

    @__call__.register
    def _(self, layer: HQuantize):
        kbiRS = self.get_kbiRS(layer)
        return FixedPointQuantizer(*kbiRS, name=layer.name)

    @__call__.register
    def _(self, layer: Signature):
        return FixedPointQuantizer(layer.keep_negative, layer.bits, layer.int_bits, 'TRN', self.SAT, name=layer.name)

    @__call__.register
    def _(self, layer: HLayerBase):
        # HLayer = Layer + Quantizer, or Layer + Quantizer1 + ReLU + Quantizer2

        name = layer.name
        klayer = to_keras_layer(layer)

        input_shape = layer.input_shape
        if isinstance(input_shape, tuple):
            inputs = keras.layers.Input(shape=input_shape[1:])
        else:
            inputs = [keras.layers.Input(shape=shape[1:]) for shape in input_shape]

        overrides = {'layers': {}}
        if hasattr(layer, 'parallel_factor') and layer.parallel_factor is not None:
            parallel_factor = layer.parallel_factor
            overrides = {'layers': {name: {'parallelization_factor': int(parallel_factor)}}}

        if not isinstance(klayer, keras.layers.Activation):
            if layer._relu_act:  # a non-activation layer can only have relu or liner as activation
                k, i, f = layer.paq.get_bits_exact(pos_only=False)
                k, i, f = tf.reduce_max(k), tf.reduce_max(i), tf.reduce_max(f)
                k, b, i = k, k + i + f, k + i
                rk, rb, ri, R, S = self.get_kbiRS(layer, pos_only=True)

                f_add_bits = 0 if R == 'TRN' else 1 if R == 'RND' else 2
                fq1 = FixedPointQuantizer(k, b + f_add_bits, i, SAT=S, RND='TRN', name=f'{name}_quantizer', overrides=overrides)
                fq2 = FixedPointQuantizer(rk, rb, ri, SAT=S, RND=R, name=f'{name}_relu_quantizer')
                return keras.Model(inputs, fq2(keras.layers.ReLU()(fq1(klayer(inputs)))))

        k, b, i, R, S = self.get_kbiRS(layer)
        q = FixedPointQuantizer(k, b, i, R, SAT=S, name=f'{name}_quantizer', overrides=overrides)
        return keras.Model(inputs, q(klayer(inputs)))

    @__call__.register
    def _(self, layer: PLayerBase):
        klayer = to_keras_layer(layer)
        return klayer

    @__call__.register
    def _(self, layer: keras.layers.Dropout):
        # Dropout does nothing at inference time
        return None

    @__call__.register
    def _(self, layer: keras.layers.Activation):
        # Purge linear activation on sight
        if layer.activation is keras.activations.linear:
            return None
        return layer


def to_proxy_model(model: keras.Model, aggressive: bool = True, accum_fp_max_offset: int | None = None, unary_lut_max_table_size=-1):
    """Given a HGQ model, return a hls4ml-ready keras model.

    Args:
        model: The HGQ model to be converted.

        aggressive (default: True): If True, use WRAP overflow mode. Significant performance degradation may occur if overflow occurs, but latency may be reduced. If False, use SAT overflow mode. Performance is more stable when it overflows, but latency may be increased.

        accum_fp_max_offset (default: None): If not None, autoset accumulator such that the model is bit accurate (when no overflow occurs and up to fp precision). If set, use the specified number of floating bits plus result float bits as accumulator float bits. May improve latency in some rare cases, not recommended in general.

        unary_lut_max_table_size (default: -1): If greater than 0, use unary LUT for `HActivation` layers, when the required table size is less than or equal to the specified value. If set to -1, do not use unary LUT.

    """

    if accum_fp_max_offset is not None and not aggressive:
        warn('You are using bitgrowth (aggressive=False) together with bias_accum_bits set. This is not recommended. If you are sure what you are doing, ignore this warning.')
    if accum_fp_max_offset is not None and accum_fp_max_offset < 0:
        warn('You are using a negative value for bias_accum_bits. Please make sure you know what you are doing.')
    proxy = convert_model(model, layer_xformer=ProxyLayerXFormer('WRAP' if aggressive else 'SAT').__call__)
    if unary_lut_max_table_size > 0:
        proxy = convert_model(proxy, layer_xformer=partial(xfr_to_unary_lut, max_table_size=unary_lut_max_table_size))
    for layer in proxy.layers:
        register_qconf(layer, accum_fp_max_offset=accum_fp_max_offset)
    return proxy



from .base import HLayerBase
from .conv import HConv, HConv1D, HConv1DBatchNorm, HConv2D, HConv2DBatchNorm
from .dense import HDense, HDenseBatchNorm
from .misc import HActivation, HAdd, HQuantize
from .passive_layers import *



from collections.abc import Callable
from functools import singledispatch

import numpy as np
import tensorflow as tf
from keras.layers import AvgPool1D, AvgPool2D, AvgPool3D, Concatenate, Flatten, MaxPool1D, MaxPool2D, MaxPool3D, Permute, Reshape
from keras.src.layers.convolutional.base_conv import Conv
from keras.src.layers.pooling.base_pooling1d import Pooling1D
from keras.src.layers.pooling.base_pooling2d import Pooling2D
from keras.src.layers.pooling.base_pooling3d import Pooling3D
from tensorflow import keras

from ..utils import apf_to_tuple, tuple_to_apf, warn
from .fixed_point_quantizer import FixedPointQuantizer

STREAM = False
"This variable is not used for now."


def get_arr_container(arr: np.ndarray, silent=False):
    """ Get the minimal fixed integer that can represent the array (kif format). If the result is greater than ~30, consider that as inf. (Not representable by fixed point with reasonable bitwidth.)"""
    k = arr < 0
    lf, hf = -32, 32
    if np.all(arr == 0):
        warn('All zero array is detected.')
        return 0, 1, 0  # All zero array is special. Though, (u)fixed<0,...> will lead to a crash, thus ufixed<1,...> is used.
    while True:
        if hf - lf <= 1:
            break
        f = (lf + hf) // 2
        _arr = arr * 2**f
        if np.all(_arr.astype(np.int64) == _arr):
            hf = f
        else:
            lf = f
    f = int(hf)
    if hf == 32 and not silent:
        warn('Failed to derive fractional bits. Is the array really quantized with less than 32 fp bits?')
    with np.errstate(divide='ignore'):
        i1, i2 = -np.inf, -np.inf
        if (~k).any():
            i1 = np.floor(1 + np.log2(arr[~k] * 2.**f)).max()
        if k.any():
            i2 = np.ceil(np.log2(-arr[k] * 2.**f)).max()
        i = int(max(i1, i2))
    i -= f
    k = int(k.any())
    return k, i, f


def activation_kif_forward(func: Callable, k: int, i: int, f: int):
    """Given the input bitwidth (kif) of an activation function, get the output bitwidth (kif)."""
    assert k + i + f > 0, 'Activation function is applied to an zero array. Something is wrong.'
    N = min(2**(k + i + f), 2**20)
    arr = np.array(np.linspace(-2.**i * k, 2.**i - 2.**-f, N), dtype=np.float64)
    arr: np.ndarray = np.array(func(arr), dtype=np.float32)
    K, I, F = get_arr_container(arr, silent=True)
    return K, I, F


def get_input_kifs(layer: keras.layers.Layer) -> tuple[tuple[int, int, int] | np.ndarray, ...]:
    """Get the input bitwidth of a layer, as a tuple of (k, i, f)."""
    parents: keras.layers.Layer | list[keras.layers.Layer] = layer._inbound_nodes[0].inbound_layers

    # As any layer that *changes* bitwidth will be followed by a quantizer, assume all layers we will meet here are "passive" layers.
    if isinstance(parents, keras.layers.Layer):
        return (get_produced_kif(parents),)
    else:
        # Passive layers with multiple inputs. Likely only Concatenate layer.
        return tuple(get_produced_kif(parent) for parent in parents)


# ===================================================================================================
# Given a layer, get the produced bitwidth of the layer, as a tuple of (k, i, f).
# ===================================================================================================

@singledispatch
def get_produced_kif(layer) -> tuple[int, int, int]:
    """Get the produced bitwidth of a layer, as a tuple of (k, i, f)."""
    warn(f'Layer {layer.name} is unknown. Assuming infinite produced bitwidth.')
    return 1, 65535, 65535


@get_produced_kif.register
def _(layer: FixedPointQuantizer):
    kifs = get_input_kifs(layer)
    assert len(kifs) == 1, f'Quantizer {layer.name} has more than one input. This is not supported.'
    _ik, _ii, _if = kifs[0]

    # If the quantizer bit width is unnecessarily large, reduce it.
    ok, oi, of = layer.result_t_kif
    k, i, f = min(_ik, ok), min(_ii, oi), min(_if, of)
    if _ii <= oi and _if > f:
        # May overflow otherwise, like (8,4) & (8,5) should result in (9,4) instead of (8,4) to avoid overflow.
        i += 1
    if k != ok or i != oi or f != of:
        layer.keep_negative.assign(tf.minimum(layer.keep_negative, k))
        layer.integers.assign(tf.minimum(layer.integers, k + i))
        layer.bits.assign(tf.minimum(layer.bits, k + i + f))

    return k, i, f


@get_produced_kif.register
def _(layer: keras.layers.Activation | keras.layers.ReLU | keras.layers.LeakyReLU | keras.layers.Softmax):
    kifs = get_input_kifs(layer)
    assert len(kifs) == 1, f'Activation layer {layer.name} has more than one input. This is not supported.'
    if isinstance(layer, keras.layers.Activation):
        if layer.activation is tf.keras.activations.softmax:
            return 0, 1, 65535
    if isinstance(layer, keras.layers.Softmax):
        return 0, 1, 65535
    k, i, f = activation_kif_forward(layer, *np.max(kifs, axis=0))  # type: ignore
    return k, i, f


@get_produced_kif.register
def _(layer: AvgPool1D | AvgPool2D | AvgPool3D):
    kifs = get_input_kifs(layer)
    assert len(kifs) == 1, f'Pooling layer {layer.name} has more than one input. This is not supported.'
    k, i, f = kifs[0]
    pool_size = np.prod(layer.pool_size)
    f += int(np.ceil(np.log2(pool_size)))
    return k, i, f


@get_produced_kif.register
def _(layer: keras.layers.Add):
    kifs = get_input_kifs(layer)
    k, i, f = np.max(kifs, axis=0)  # type: ignore
    # being lazy here. But this will never overflow.
    i += int(np.ceil(np.log2(len(kifs))))
    return k, i, f


@get_produced_kif.register
def _(layer: keras.layers.Concatenate):
    kifs = get_input_kifs(layer)
    k, i, f = np.max(kifs, axis=0)  # type: ignore
    return k, i, f


@get_produced_kif.register
def _(layer: keras.layers.Dense | Conv):
    kifs = get_input_kifs(layer)
    assert len(kifs) == 1, f'Dense layer {layer.name} has more than one input. This is not supported.'
    k, i, f = kifs[0]
    w_k, w_i, w_f = get_arr_container(layer.kernel.numpy())
    k = int(k or w_k or np.any(layer.use_bias and layer.bias < 0))  # type: ignore
    i, f = i + w_i, f + w_f
    if isinstance(layer, Conv):
        div = layer.filters
    else:
        div = layer.units

    multiplicity = np.prod(layer.kernel.shape) / div
    i += int(np.ceil(np.log2(multiplicity)))
    return k, i, f


@get_produced_kif.register
def _(layer: Reshape | Flatten | MaxPool3D | MaxPool2D | MaxPool1D | Permute):
    kifs = get_input_kifs(layer)
    assert len(kifs) == 1, f'Flatten/Reshape layer {layer.name} has more than one input. This is not supported.'
    k, i, f = kifs[0]
    return k, i, f


@get_produced_kif.register
def _(layer: keras.layers.InputLayer):
    return 1, 65535, 65535


# ===================================================================================================


def get_requested_kif(layer: keras.layers.Layer | FixedPointQuantizer) -> tuple[int, int, int]:
    """Get the bitwidth requested by downstream layers, as a tuple of (k, i, f). By requested, it means the maximum bitwidth that downstream layers may make use of."""
    out_layers = [node.outbound_layer for node in layer._outbound_nodes]
    if len(out_layers) == 0:
        # Terminal case
        return (1, 65535, 65535)
    else:
        # Main case.
        requested_kifs = [get_request_kif(out_layer) for out_layer in out_layers]
        k, i, f = np.max(requested_kifs, axis=0)
        return k, i, f

# ===================================================================================================
# Given a layer, get the bitwidth it requests (maximum bitwidth it can make use of), as a tuple of (k, i, f).
# ===================================================================================================


@singledispatch
def get_request_kif(layer: keras.layers.Layer) -> tuple[int, int, int]:
    """Get the requested bitwidth of a layer, as a tuple of (k, i, f)"""
    if isinstance(layer, (Concatenate, Reshape, Flatten)):
        out_layers: list[keras.layers.Layer] = [node.outbound_layer for node in layer._outbound_nodes]
        if out_layers:
            # Layers that does nothing. Pass through.
            requested_kifs = [get_request_kif(out_layer) for out_layer in out_layers]
            k, i, f = np.max(requested_kifs, axis=0)
            return k, i, f
    return 1, 65535, 65535


@get_request_kif.register
def _(layer: FixedPointQuantizer):
    k, i, f = layer.result_t_kif
    rnd = layer.RND
    sat = layer.SAT
    if rnd.upper() == 'TRN':
        f += 0
    elif rnd.upper() == 'RND':
        f += 1
    else:
        f += 2
    if 'WRAP' not in sat.upper():
        i = 65535
    return k, i, f


# ===================================================================================================


def merge_precision(available: tuple[int, int, int], request: tuple[int, int, int]):
    """Given available precision and the maximum precision can be accepted by downstream, return the precision should be allocated for the data path."""
    k0, i0, f0 = available
    k1, i1, f1 = request

    k = k0
    i = min(i0, i1)
    f = min(f0, f1)
    if f0 > f1 and i0 <= i1:
        i += 1
    return k, i, f


def result_kifRS_layer_with_fusible_quantizer(layer: keras.layers.Layer):
    """Get the result bitwidth of a layer that has a fusible quantizer following immediately, as a tuple of (k, i, f, RND, SAT). When the layer has exactly one quantizer following it, and the quantizer is not heterogenous, the quantizer will be purged during synthesis, and the result bitwidth of the layer will be the same as the quantizer."""
    out_nodes = layer._outbound_nodes
    if len(out_nodes) != 1:
        return
    out_layer = out_nodes[0].outbound_layer
    if not isinstance(out_layer, FixedPointQuantizer):
        return
    if not out_layer.fusible:
        return
    k, i, f = out_layer.result_t_kif
    SAT = out_layer.SAT
    RND = out_layer.RND
    return k, i, f, RND, SAT


def derive_result_kifRS_from_next_quantizers(layer: keras.layers.Layer) -> tuple[int, int, int, str, str]:
    """Get the result bitwidth of a layer that has a quantizer following immediately, as a tuple of (k, i, f, RND, SAT). In general, any InputLayer or layers with kernels will have a quantizer following immediately."""
    Ks, Is, Fs, RNDs, SATs = [], [], [], [], []
    out_layers: list[FixedPointQuantizer] = [node.outbound_layer for node in layer._outbound_nodes]
    for out_layer in out_layers:
        assert isinstance(out_layer, FixedPointQuantizer), f'Layer {layer.name} has a non-quantizer layer {out_layer.name} following it. This is not supported.'
    kifRSs = np.array(list(map(lambda x: (*x.result_t_kif, x.RND, x.SAT), out_layers)), dtype=object)
    Ks, Is, Fs, RNDs, SATs = kifRSs.T
    assert np.all(SATs == SATs[0]), f'Input layer {layer.name} has different SATs. This is not supported.'
    SAT = SATs[0].upper()
    if SAT == 'WRAP_SM':
        warn('WRAP_SM for input layer may lead bitmismatch between proxy and hls4ml due to overflow on th first layer.')
    k = np.max(Ks)
    i = np.max(Is)
    f = np.max(Fs + 2 * (RNDs != 'TRN') - 1 * (RNDs == 'RND'))
    return k, i, f, 'TRN', SAT


def get_result_kifRS(layer: keras.layers.Layer) -> tuple[int, int, int, str, str]:
    """Get the result bitwidth of a layer, as a tuple of (k, i, f, RND, SAT)."""
    result_t = result_kifRS_layer_with_fusible_quantizer(layer)
    if result_t is not None:
        return result_t
    if isinstance(layer, keras.layers.InputLayer):
        return derive_result_kifRS_from_next_quantizers(layer)
    if STREAM and (hasattr(layer, 'kernel')):
        return derive_result_kifRS_from_next_quantizers(layer)

    produced_kif = get_produced_kif(layer)
    requested_kif = get_requested_kif(layer)
    k, i, f = merge_precision(produced_kif, requested_kif)
    return k, i, f, 'TRN', 'WRAP'


def get_config_wight_accum_result_bias(layer: keras.layers.Layer, accum_fp_max_offset: None | int = None):
    """Get the quantization configuration for a layer with kernel in the proxy model."""
    assert hasattr(layer, 'kernel'), f'Layer {layer.name} does not have kernel.'
    r_k, r_i, r_f, RND, SAT = get_result_kifRS(layer)
    p_k, p_i, p_f = get_produced_kif(layer)
    k_k, k_i, k_f = get_arr_container(layer.kernel.numpy())
    weight_t = tuple_to_apf((k_k, k_i, k_f))
    result_t = tuple_to_apf((r_k, r_i, r_f), RND, SAT)
    if accum_fp_max_offset is None:
        a_f = p_f
    else:
        a_f = min(r_f + accum_fp_max_offset, p_f)

    if SAT.upper() == 'WRAP':
        accum_t = tuple_to_apf((r_k, r_i, a_f))
    else:
        k, i, f = p_k, p_i, a_f
        if accum_fp_max_offset is not None:
            warn(f'Layer {layer.name} has SAT={SAT}, and has accum_fp_max_offset set. Proceed only if you know what you are doing.')
            f = min(accum_fp_max_offset + r_f, f)
        accum_t = tuple_to_apf((k, i, f))
    bias_t = accum_t
    return weight_t, accum_t, result_t, bias_t


def get_config_table_tablesize_result(layer: keras.layers.Activation):
    """Get the quantization configuration for a activation layer in the proxy model."""
    assert isinstance(layer, keras.layers.Activation), f'Layer {layer.name} is not an activation layer.'
    k, i, f, RND, SAT = get_result_kifRS(layer)
    result_t = tuple_to_apf((k, i, f), RND, SAT)
    table_t = result_t
    if layer.activation in (keras.activations.tanh, keras.activations.sigmoid):
        result_t = tuple_to_apf((k, i, f), 'TRN', 'WRAP')  # These activations are purely LUT based, this rounding/saturating in the table only would be enough.
    i_kifs = get_input_kifs(layer)
    if len(i_kifs) > 1:
        warn(f'Activation layer {layer.name} has more than one input. Did you just make a activation with multiple inputs? table_size set in this way may make no sense. Proceed only if you know what you are doing.')
    i_k, i_i, i_f = np.max(i_kifs, axis=0)  # type: ignore
    table_size = int(2**(i_k + i_i + i_f))  # ...the ideal case. Will be the case if we have universal LUT-based activation.
    if layer.activation is keras.activations.tanh:
        table_size = int(8 / 2.**-i_f)  # LUT Range hardcoded to -4 ~ 4, match #fractional bits
    elif layer.activation is tf.keras.activations.sigmoid:
        table_size = int(16 / 2.**-i_f)  # LUT Range hardcoded to -8 ~ 8, match #fractional bits
    return table_t, table_size, result_t


def get_config(layer: keras.layers.Layer, accum_fp_max_offset: None | int = None):
    """Get the quantization configuration for a layer in the proxy model."""
    if hasattr(layer, 'kernel'):
        weight_t, accum_t, result_t, bias_t = get_config_wight_accum_result_bias(layer, accum_fp_max_offset)
        conf = {'weight_t': weight_t, 'accum_t': accum_t, 'result_t': result_t, 'bias_t': bias_t}
    elif isinstance(layer, keras.layers.Activation):
        table_t, table_size, result_t = get_config_table_tablesize_result(layer)
        conf = {'table_t': table_t, 'table_size': table_size, 'result_t': result_t}
    else:
        k, i, f, RND, SAT = get_result_kifRS(layer)
        result_t = tuple_to_apf((k, i, f), RND, SAT)
        conf = {'result_t': result_t}
    return conf


def _get_next_quantizer(layer: keras.layers.Layer):
    """Get the next quantizer after the layer. Return None if there is no quantizer after the layer."""
    out_layers = [node.outbound_layer for node in layer._outbound_nodes]
    for out_layer in out_layers:
        if isinstance(out_layer, FixedPointQuantizer):
            return out_layer
        r = _get_next_quantizer(out_layer)
        if r is not None:
            return r
    return None


def _get_last_quantizer(layer: keras.layers.Layer):
    """Get the last quantizer before the layer. Return None if there is no quantizer before the layer."""
    assert len(layer._inbound_nodes) <= 1, f'Layer {layer.name} has more than one inbound nodes. This is not supported.'
    in_layers = layer._inbound_nodes[0].inbound_layers
    if not isinstance(in_layers, list):
        in_layers = [in_layers]
    for in_layer in in_layers:
        if isinstance(in_layer, FixedPointQuantizer):
            return in_layer
        r = _get_last_quantizer(in_layer)
        if r is not None:
            return r
    return None


def get_whatever_quantizer(layer: keras.layers.Layer):
    """Find the quantizer before or after the layer."""
    if isinstance(layer, FixedPointQuantizer):
        return layer
    if isinstance(layer, keras.layers.InputLayer):
        pass
    q = _get_next_quantizer(layer) or _get_last_quantizer(layer)
    assert q is not None, f'Layer {layer.name} has no quantizer before or after it. Did you create a valid proxy model?'
    return q


def register_qconf(layer: keras.layers.Layer, accum_fp_max_offset: None | int = None):
    """Get and register quantization configuration for a layer in the proxy model."""
    if hasattr(layer, 'proxy_ready'):
        return
    q = get_whatever_quantizer(layer)
    conf = get_config(layer, accum_fp_max_offset=accum_fp_max_offset)
    overrides = q.overrides or {}
    overrides['layers'].setdefault(layer.name, {}).update(conf)



from collections.abc import Callable
from functools import singledispatchmethod

import numpy as np
import tensorflow as tf

from ..utils import strategy_dict

two = tf.constant(2, dtype=tf.float32)
log2 = tf.constant(np.log(2), dtype=tf.float32)


@tf.function(jit_compile=True)
def q_round(x: tf.Tensor, strategy: int = 0):
    """Round the tensor.

    strategy:
        0: standard round (default, round to nearest, 0.5 to even)
        1: stochastic round
        2: fast uniform noise injection (uniform noise in [-0.5, 0.5])
        3: floor

    """
    if strategy == 0:  # standard round
        return tf.floor(x + 0.5)  # type: ignore
    if strategy == 1:  # stochastic round
        _floor = tf.floor(x)
        noise = tf.random.uniform(tf.shape(x))
        return tf.where(noise < x - _floor, _floor + 1, _floor)
    if strategy == 2:  # fast uniform noise injection
        noise = tf.random.uniform(tf.shape(x), minval=-0.5, maxval=0.5)  # type: ignore
        noise = tf.where(tf.abs(x) <= 0.5, -x, noise)  # type: ignore
        return tf.stop_gradient(noise) + x  # type: ignore
    if strategy == 3:
        return tf.floor(x)
    raise ValueError(f"Unknown strategy {strategy}")


def get_arr_bits(arr: np.ndarray):
    """Internal helper function to compute the position of the highest and lowest bit of an array of fixed-point integers."""
    kn = (arr < 0).astype(np.int8)
    mul = 32  # type: ignore
    arr = arr * 2**mul
    arr = np.abs(arr)[..., None]  # type: ignore
    n = int(np.ceil(np.max(np.log2(arr + 1))))  # type: ignore
    divisor = 2**np.arange(1, n)[None, ...]  # type: ignore
    low_pos = np.sum(arr % divisor == 0, axis=-1) + (arr[..., 0] == 1)
    with np.errstate(divide='ignore'):
        high_pos = np.where(arr[..., 0] != 0, np.floor(np.log2(arr[..., 0]) + 1), low_pos).astype(np.int8)
    fb = 32 - low_pos
    int_bits = high_pos - low_pos - fb
    zero_mask = int_bits + fb == 0
    int_bits[zero_mask] = 0
    fb[zero_mask] = 0
    return kn.astype(np.int8), int_bits.astype(np.int8), fb.astype(np.int8)


class HGQ:
    """Heterogenous quantizer."""

    def __init__(self, init_bw: float, skip_dims, rnd_strategy: str | int = 'floor', exact_q_value=True, dtype=None, bw_clip=(-23, 23), trainable=True, regularizer: Callable | None = None, minmax_record=False):
        self.init_bw = init_bw
        self.skip_dims = skip_dims
        """tuple: Dimensions to use uniform quantizer. If None, use full heterogenous quantizer."""
        self.rnd_strategy = strategy_dict.get(rnd_strategy, -1) if isinstance(rnd_strategy, str) else rnd_strategy
        """How to round the quantized value. 0: standard round (default, round to nearest, round-up 0.5), 1: stochastic round, 2: fast uniform noise injection (uniform noise in [-0.5, 0.5]), 3: floor"""
        self.exact_q_value = exact_q_value
        """bool: Whether to use exact quantized value during training."""
        self.dtype = dtype
        self.bw_clip = bw_clip
        """tuple: (min, max) of bw. 23 by default in favor of float32 mantissa."""
        self.trainable = trainable
        self.regularizer = regularizer
        """Regularizer for bw."""
        self.minmax_record = minmax_record
        """bool: Whether to record min and max of quantized values."""
        self.built = False
        self.degeneracy = 1.
        """Degeneracy of the quantizer. Records how many values are mapped to the same quantizer."""

    def _compute_bw_shape_and_degeneracy(self, input_shape):
        """Map skip_dims to input_shape and compute degeneracy."""
        if isinstance(self.skip_dims, str):
            if self.skip_dims == 'all':
                self.skip_dims = tuple(range(len(input_shape)))
            elif self.skip_dims == 'batch':
                self.skip_dims = (0,)
            elif self.skip_dims == 'none':
                self.skip_dims = None
            elif self.skip_dims == 'except_last':
                self.skip_dims = tuple(range(len(input_shape) - 1))
            elif self.skip_dims == 'except_1st':
                self.skip_dims = (0,) + tuple(range(2, len(input_shape)))
            else:
                raise ValueError("skip_dims must be tuple or str in ['all', 'except_last', 'batch', 'except_last', 'except_1st', 'none']")
        _input_shape = list(input_shape)
        degeneracy = 1
        if self.skip_dims:
            for d in self.skip_dims:
                degeneracy *= int(_input_shape[d]) if _input_shape[d] is not None else 1
                _input_shape[d] = 1
        return _input_shape, degeneracy

    def init_minmax(self):
        if self.minmax_record:
            dtype = self.dtype or tf.keras.backend.floatx()
            self._min = tf.Variable(tf.zeros_like(self.fbw), trainable=False, name='min', dtype=dtype)
            self._max = tf.Variable(tf.zeros_like(self.fbw), trainable=False, name='max', dtype=dtype)
            self.minmax_reg_reset()  # type: ignore

    @singledispatchmethod
    def build(self, x, name=None):
        self.built = True
        self.fbw = x
        self.init_minmax()

    @build.register
    def _(self, input_shape: tuple, name: str | None = None):
        self.built = True
        _input_shape, degeneracy = self._compute_bw_shape_and_degeneracy(input_shape)
        self.degeneracy = degeneracy

        self.fbw = tf.Variable(
            tf.ones(_input_shape) * self.init_bw, trainable=self.trainable, name=name, dtype=tf.float32
        )

        self.init_minmax()

    @tf.function(jit_compile=True)
    def minmax_reg_reset(self):
        """Reset min and max to inf and -inf, respectively."""
        assert self.built
        inf = tf.ones(self.fbw.shape) * float('inf')
        self._min.assign(inf)
        self._max.assign(-inf)

    @tf.function(jit_compile=True)
    def __call__(self, x, training=None, record_minmax=None):
        if not self.built:
            self.build(tuple(x.shape), name=None)
        if self.bw_clip:
            self.fbw.assign(tf.clip_by_value(self.fbw, *self.bw_clip))
        return self.forward(x, training, record_minmax)  # type: ignore

    @tf.function(jit_compile=True)
    def forward(self, x, training=None, record_minmax=None):
        """Forward pass of HGQ.
        Args:
            training: if set to True, gradient will be propagated through the quantization process.
            record_minmax: if set to True, min and max of quantized values will be recorded for deriving the necessary integer bits. Only necessary for activation/pre-activation values.
        """
        if self.exact_q_value or not training:
            scale = tf.pow(two, tf.round(self.fbw))
        else:
            scale = tf.pow(two, self.fbw)

        rnd_strategy = self.rnd_strategy
        if not training and rnd_strategy != 3:  # not std round or floor
            rnd_strategy = 0
        xq = q_round(x * scale, rnd_strategy) / scale  # type: ignore

        delta = tf.stop_gradient(xq - x)
        if training:
            prod = delta * self.fbw * log2  # type: ignore
            delta = tf.stop_gradient(delta + prod) - prod

        if not record_minmax:
            return x + delta
        xq = x + delta
        if self.skip_dims:
            min_xq = tf.reduce_min(xq, axis=self.skip_dims)
            max_xq = tf.reduce_max(xq, axis=self.skip_dims)
        else:
            min_xq = max_xq = xq

        self._min.assign(tf.minimum(min_xq, self._min))
        self._max.assign(tf.maximum(max_xq, self._max))

        return xq

    @tf.function(jit_compile=True)
    def bias_forward(self, x, training=None, channel_loc=-1):
        """Forward pass for the bias term. Grammatical sugar"""
        if channel_loc == -1:
            dims = list(range(len(self.fbw.shape) - 1))
        elif channel_loc == 1:
            dims = [0] + list(range(2, len(self.fbw.shape)))
        else:
            raise ValueError('channel_loc must be -1 or 1')

        fbw = tf.reduce_max(self.fbw, axis=dims, keepdims=False)

        if self.exact_q_value or not training:
            scale = tf.pow(two, tf.round(fbw))
        else:
            scale = tf.pow(two, fbw)

        rnd_strategy = self.rnd_strategy
        if not training and rnd_strategy != 3:  # not std round or floor
            rnd_strategy = 0
        xq = q_round(x * scale, rnd_strategy) / scale  # type: ignore

        delta = tf.stop_gradient(xq - x)

        if training:
            prod = delta * fbw * log2  # type: ignore
            delta = tf.stop_gradient(delta + prod) - prod

        return x + delta

    @tf.function(jit_compile=True)
    def get_bits(self, ref=None, quantized=None, pos_only=False):
        """Get approximated int/frac/keep_negative bits of the equivalent fixed-point quantizer.
        Args:
            ref: Input tensor to compute the bits. If None, use the min/max record.
            quantized: If input is already quantized. Skip quantization pass if set to True.
            pos_only: If True, only compute the bits for positive values. Useful if have a ReLU layer after.
        """
        fp_bits = tf.round(self.fbw)
        fp_bits = self.fbw + tf.stop_gradient(fp_bits - self.fbw)  # type: ignore
        if ref is not None:
            if quantized is not None:
                _ref = ref
            else:
                _ref = self.forward(ref)  # type: ignore
            kn = tf.keras.backend.cast_to_floatx(_ref < 0)
            _ref = tf.abs(_ref)
            int_bits = tf.floor(tf.math.log(_ref) / log2) + 1
            if self.skip_dims:
                int_bits = tf.reduce_max(int_bits, axis=self.skip_dims)
                kn = tf.reduce_max(kn, axis=self.skip_dims)
        else:
            assert self.minmax_record
            if pos_only:
                _ref = tf.maximum(self._max, 0.)
                kn = tf.zeros_like(self._max)
            else:
                _ref = tf.maximum(tf.abs(self._min), tf.abs(self._max))
                kn = tf.keras.backend.cast_to_floatx(self._min < 0)  # type: ignore
            int_bits = tf.floor(tf.math.log(_ref) / log2) + 1
        return int_bits, fp_bits, kn

    def get_bits_exact(self, ref=None, pos_only=False):
        """Get exact int/frac/keep_negative bits of the equivalent fixed-point quantizer.
        Args:
            ref: Input tensor to compute the bits. If None, use the min/max record.
            pos_only: If True, only compute the bits for positive values. Useful if have a ReLU layer after.
        """

        if ref is None and self.minmax_record:
            assert tf.reduce_all(self._max - self._min >= 0), "Some min values are larger than max values. Did you forget to run trace_minmax?"  # type: ignore
            f = tf.round(self.fbw).numpy()  # type: ignore
            with np.errstate(divide='ignore'):
                if pos_only:
                    _ref = np.maximum(self._max, 0.)  # type:ignore
                    i = np.floor(np.log2(_ref)) + 1
                    k = np.zeros_like(self._max)
                else:
                    i_neg = np.ceil(np.log2(np.abs(self._min)))  # type:ignore
                    i_pos = np.ceil(np.log2(np.abs(self._max + 2.**-f)))  # type:ignore
                    i = np.maximum(i_neg, i_pos)
                    k = (self._min.numpy() < 0)  # type:ignore
            i = np.clip(i, -f - k, 32)
            k, i, f = k.astype(np.int8), i.astype(np.int8), f.astype(np.int8)
            mask = k + i + f != 0
            return k * mask, i * mask, f * mask

        assert ref is not None
        w = self.forward(ref).numpy()  # type: ignore
        k, i, f = get_arr_bits(w)
        mask = k + i + f != 0
        return k * mask, i * mask, f * mask

    def adapt_bw_bits(self, ref: tf.Tensor):
        """Adapt the bitwidth of the quantizer to the input tensor, such that each input is represented with approximately the same number of bits. (i.e., 1.5 with be represented by ap_fixed<2,1> and 0.375 will be represented by ap_fixed<2,-2>)."""
        if not self.built:
            self.build(tuple(ref.shape), name=None)
        new_fbw = self.fbw - (tf.math.log(tf.abs(ref)) / log2)
        if self.skip_dims:
            new_fbw = tf.reduce_min(new_fbw, axis=self.skip_dims, keepdims=True)
        self.fbw.assign(new_fbw)

    @classmethod
    def from_config(cls, config: dict):
        return cls(**config)



import abc

import keras
import numpy as np
import qkeras
from keras import backend as K
from qkeras import quantizers
from qkeras.quantizers import BaseQuantizer
from qkeras.utils import _add_supported_quantized_objects
from tensorflow import keras

from HGQ.proxy.fixed_point_quantizer import FixedPointQuantizer
from HGQ.utils import warn

from ..convert import ProxyLayerXFormer
from ..precision_derivation import get_produced_kif

qkeras_objects = {}
_add_supported_quantized_objects(qkeras_objects)

qkeras_layers = {v for v in qkeras_objects.values() if issubclass(v, keras.layers.Layer)}
qkeras_quantizers = {v for v in qkeras_objects.values() if issubclass(v, BaseQuantizer)}
real_qkeras_quantizers = {
    quantizers.quantized_bits,
    quantizers.quantized_relu,
    quantizers.stochastic_binary,
    quantizers.stochastic_ternary,
    quantizers.ternary,
    quantizers.binary,
    quantizers.quantized_po2,
    quantizers.quantized_relu_po2,
}

from functools import singledispatch


@singledispatch
def qkeras_quantizer_to_layers(quantizer, SAT) -> tuple[keras.layers.Layer, ...]:
    """
    Map qkeras quantizer to a tuple of equivalent proxy layers.
        e.g.: quantized_bits -> (fixed,)
              quantized_relu -> (fixed, relu, fixed)
    """
    if quantizer is keras.activations.linear:
        return ()
    return keras.layers.Activation(quantizer),


@qkeras_quantizer_to_layers.register
def _(quantizer: quantizers.quantized_bits, SAT):
    assert quantizer.alpha == 1, f"alpha != 1 is not currently supported, got {quantizer.alpha}."
    k = quantizer.keep_negative
    i = quantizer.integer + k
    b = quantizer.bits
    SYM = bool(quantizer.symmetric)
    if SYM:
        if SAT != 'SAT':
            warn(f'Symmetric quantizer is only possible with SAT=SAT_SYM, but got SAT={SAT}. Enforcing SAT=SAT.')
        SAT = 'SAT_SYM'
    if k == b == 1:
        warn(f"Qkeras's implementation for quantizers with keep_negative = bits = 1 is flawed as of 2023/11. If you believe they have fixed it, please ignore this warning.")
    if quantizer.symmetric != 0 and quantizer.symmetric != 1:
        raise ValueError(f"Expecting symmetric to be 0 or 1 for quantized_bits, got {quantizer.symmetric}.")

    return FixedPointQuantizer(k, b, i, 'RND_CONV', SAT),


@qkeras_quantizer_to_layers.register
def _(quantizer: quantizers.quantized_relu, SAT):
    i = quantizer.integer
    b = quantizer.bits
    q0 = FixedPointQuantizer(1, b + 1, i + 1, 'RND_CONV', SAT)
    q1 = FixedPointQuantizer(0, b, i, 'RND_CONV', 'WRAP')
    relu = keras.layers.Activation('relu')
    return q0, relu, q1


@qkeras_quantizer_to_layers.register
def _(quantizer: quantizers.stochastic_binary | quantizers.binary, SAT):
    if quantizer.use_01:
        raise NotImplementedError("No support is currently available for use_01=True in stochastic_binary or binary on hls4ml.")
        return FixedPointQuantizer(0, 1, 1, 'RND', 'SAT'),
    else:
        return qkeras.QActivation(qkeras.binary()), FixedPointQuantizer(1, 2, 2, 'RND', 'WRAP')


@qkeras_quantizer_to_layers.register
def _(quantizer: quantizers.stochastic_ternary | quantizers.ternary):
    return FixedPointQuantizer(1, 2, 1, 'RND', 'SAT')


@qkeras_quantizer_to_layers.register
def _(quantizer: quantizers.quantized_po2 | quantizers.quantized_relu_po2):
    raise ValueError("quantized_po2 family cannot be implemented in hls4ml as activation.")


def qlayer_to_keras_layer(layer: keras.layers.Layer) -> keras.layers.Layer | None:
    base_cls_name = layer.__class__.__name__[1:]
    if not hasattr(keras.layers, base_cls_name):
        raise ValueError(f"Cannot find corresponding keras layer for {layer.__class__.__name__}")
    base_cls = getattr(keras.layers, base_cls_name)
    conf = layer.get_config()

    # For some ridiculous reason, QAactivation's activation is called quantizer
    # Yet, all other layers are call their activation activation.
    if isinstance(layer, (qkeras.QActivation, keras.layers.Activation)):
        quantizer = layer.quantizer if isinstance(layer, qkeras.QActivation) else layer.activation
        if quantizer.__class__ in qkeras_quantizers:
            return None

    blacklist = ('quantize', 'kernel_range', 'bias_range', 'constraint', 'activation', 'initializer')
    for k in list(conf.keys()):
        for k_rm in blacklist:
            if k_rm in k:
                del conf[k]
                continue

    klayer = base_cls(**conf)
    klayer.trainable = False
    klayer.build(layer.input_shape)
    if hasattr(layer, 'kernel'):
        q = layer.kernel_quantizer_internal or (lambda x: x)
        klayer.kernel.assign(q(layer.kernel))
    if hasattr(layer, 'bias') and layer.use_bias:
        q = layer.bias_quantizer_internal or (lambda x: x)
        klayer.bias.assign(q(layer.bias))
    return klayer


def qlayer_to_proxy_layer(self: ProxyLayerXFormer, layer: keras.layers.Layer) -> tuple[keras.layers.Layer, ...]:

    SAT = self.SAT
    klayer = qlayer_to_keras_layer(layer)
    if not hasattr(layer, 'activation'):
        assert klayer is not None
        return klayer

    activation = layer.quantizer if hasattr(layer, 'quantizer') else layer.activation
    act_layers = qkeras_quantizer_to_layers(activation, SAT)

    input_shape = layer.input_shape
    if isinstance(input_shape, tuple):
        inputs = keras.layers.Input(shape=input_shape[1:])
    else:
        inputs = [keras.layers.Input(shape=shape[1:]) for shape in input_shape]

    out = inputs
    if klayer is not None:
        out = klayer(out)
    for layer in act_layers:
        out = layer(out)
    return keras.Model(inputs, out)


class QKerasBaseLayer(metaclass=abc.ABCMeta):
    pass


for layer_cls in qkeras_layers:
    QKerasBaseLayer.register(layer_cls)


def init():
    ProxyLayerXFormer.__call__.register(QKerasBaseLayer, qlayer_to_proxy_layer)  # type: ignore
    get_produced_kif.register(qkeras.QActivation, get_produced_kif.registry[keras.layers.Activation])



import re
from warnings import warn as _warn

import keras
import tensorflow as tf


@keras.utils.register_keras_serializable(package='HGQ')
class MonoL1:
    def __init__(self, l1=0.):
        assert l1 >= 0, f'l1 must be non-negative, got {l1}'
        self.l1 = l1

    def __call__(self, x):
        return tf.reduce_sum(x) * self.l1

    def get_config(self):
        return {'l1': self.l1}


DEFAULT_KERNEL_QUANTIZER_CONFIG = \
    dict(init_bw=2,
         skip_dims=None,
         rnd_strategy='standard_round',
         exact_q_value=True,
         dtype=None,
         bw_clip=(-23, 23),
         trainable=True,
         regularizer=MonoL1(1e-6),
         )


DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG = \
    dict(init_bw=2,
         skip_dims=(0,),
         rnd_strategy='standard_round',  # 'auto': 'floor' for layer without bias or with 'relu' activation, 'standard_round' otherwise
         exact_q_value=True,
         dtype=None,
         bw_clip=(-23, 23),
         trainable=True,
         regularizer=MonoL1(1e-6),
         minmax_record=True
         )


def get_default_kq_conf():
    return DEFAULT_KERNEL_QUANTIZER_CONFIG.copy()


def get_default_paq_conf():
    return DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG.copy()


def set_default_kq_conf(config):
    global DEFAULT_KERNEL_QUANTIZER_CONFIG
    assert isinstance(config, dict), f'config must be a dict, got {config.__class__.__name__}'
    assert set(config.keys()) == set(DEFAULT_KERNEL_QUANTIZER_CONFIG.keys()), \
        f'config must have keys {DEFAULT_KERNEL_QUANTIZER_CONFIG.keys()}, got {config.keys()}'
    DEFAULT_KERNEL_QUANTIZER_CONFIG = config


def set_default_paq_conf(config):
    global DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG
    assert isinstance(config, dict), f'config must be a dict, got {config.__class__.__name__}'
    assert set(config.keys()) == set(DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG.keys()), \
        f'config must have keys {DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG.keys()}, got {config.keys()}'
    DEFAULT_PRE_ACTIVATION_QUANTIZER_CONFIG = config


strategy_dict = {
    'standard_round': 0,
    'stochastic_round': 1,
    'fast_uniform_noise_injection': 2,
    'floor': 3,
}

_apf_m = re.compile(r'\s*(?:|)(u?)fixed<\s*(\d+)\s*,\s*(-?\d+)[\s,_\w]*>\s*')


def apf_to_tuple(apf: str):
    """Convert fixed format to tuple of (keep_negative, int_bits, fp_bits)"""
    m = _apf_m.match(apf)
    assert m is not None, f'Unable to parse "{apf}". The format should be "()(u)fixed<m,n...>"'
    u, b, i = m.groups()
    b, i = int(b), int(i)
    kn = 0 if u else 1
    f = b - i
    i = i - kn
    return kn, i, f


def tuple_to_apf(t: tuple, rnd='TRN', warp='WARP', keep_zeros=True):
    """Convert tuple of (keep_negative, int_bits, fp_bits) to fixed format"""
    kn, i, f = t
    if not keep_zeros and i + f + kn <= 0:
        return 'nuke'
    if rnd != 'TRN' and warp == 'WARP':
        return f'{"u" if kn==0 else ""}fixed<{max(i+f+kn,1)},{i+kn},{rnd}>'
    if warp != 'WARP':
        return f'{"u" if kn==0 else ""}fixed<{max(i+f+kn,1)},{i+kn},{rnd},{warp}>'
    return f'{"u" if kn==0 else ""}fixed<{max(i+f+kn,1)},{i+kn}>'


def warn(msg: str, stacklevel=2):
    # print in yellow
    _warn(f'\033[93m{msg}\033[0m', stacklevel=stacklevel)



import numpy as np
import pytest
import tensorflow as tf
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras

import HGQ
from HGQ import get_default_paq_conf, set_default_paq_conf
from HGQ.layers import HConv1D, HDense, HQuantize, PAvgPool1D, PAvgPool2D, PConcatenate, PFlatten, PMaxPool1D, PMaxPool2D, PReshape, Signature
from HGQ.proxy.fixed_point_quantizer import gfixed


def create_model(layer: str, rnd_strategy: str, io_type: str):
    pa_config = get_default_paq_conf()
    pa_config['rnd_strategy'] = rnd_strategy
    pa_config['skip_dims'] = 'all' if io_type == 'io_stream' else 'batch'
    set_default_paq_conf(pa_config)

    inp = keras.Input(shape=(15))
    if 'PConcatenate' in layer:
        _inp = [HQuantize()(inp)] * 2
        out = eval(layer)(_inp)
        out = HDense(15)(out)
        return keras.Model(inp, out)
    elif 'Signature' in layer:
        _inp = eval(layer)(inp)
        out = HDense(15)(_inp)
        return keras.Model(inp, out)
    elif 'Pool2D' in layer:
        _inp = PReshape((3, 5, 1))(HQuantize()(inp))
    elif 'Pool1D' in layer:
        _inp = PReshape((5, 3))(HQuantize()(inp))
    elif 'Dense' in layer or 'Activation' in layer:
        _inp = HQuantize()(inp)
    elif 'Flatten' in layer:
        out = HQuantize()(inp)
        out = PReshape((3, 5))(out)
        out = HConv1D(2, 2)(out)
        out = eval(layer)(out)
        out = HDense(15)(out)
        return keras.Model(inp, out)
    else:
        raise Exception(f'Please add test for {layer}')

    out = eval(layer)(_inp)
    if 'maxpool' in layer.lower():
        out = HQuantize()(out)
    model = keras.Model(inp, out)

    for layer in model.layers:
        # No weight bitwidths to randomize
        # And activation bitwidths
        if hasattr(layer, 'paq'):
            fbw: tf.Variable = layer.paq.fbw
            fbw.assign(tf.constant(np.random.uniform(4, 8, fbw.shape).astype(np.float32)))

    return model


def get_data(N: int, sigma: float, max_scale: float, seed):
    rng = np.random.default_rng(seed)
    a1 = rng.normal(0, sigma, (N, 15)).astype(np.float32)
    a2 = rng.uniform(0, max_scale, (1, 15)).astype(np.float32)
    return (a1 * a2).astype(np.float32)


@pytest.mark.parametrize('layer',
                         [
                             "PConcatenate()",
                             "PMaxPool1D(2, padding='same')",
                             "PMaxPool1D(4, padding='same')",
                             "PMaxPool2D((5,3), padding='same')",
                             "PMaxPool1D(2, padding='valid')",
                             "PMaxPool2D((2,3), padding='valid')",
                             "Signature(1,6,3)",
                             "PAvgPool1D(2, padding='same')",
                             "PAvgPool2D((1,2), padding='same')",
                             "PAvgPool2D((2,2), padding='same')",
                             "PAvgPool1D(2, padding='valid')",
                             "PAvgPool2D((1,2), padding='valid')",
                             "PAvgPool2D((2,2), padding='valid')",
                             "PFlatten()",
                         ]
                         )
@pytest.mark.parametrize("N", [1000])
@pytest.mark.parametrize("rnd_strategy", ['auto', 'standard_round', 'floor'])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("cover_factor", [0.25, 1.0])
@pytest.mark.parametrize("aggressive", [True, False])
@pytest.mark.parametrize("backend", ['vivado', 'vitis'])
@pytest.mark.parametrize("seed", [42])
def test_syn_players(layer, N: int, rnd_strategy: str, io_type: str, cover_factor: float, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model(layer=layer, rnd_strategy=rnd_strategy, io_type=io_type)
    data = get_data(N, 1, 1, seed)

    if 'Signature' in layer:
        q = gfixed(1, 6, 3)
        data = q(data).numpy()
    if "padding='same'" in layer and io_type == 'io_stream':
        pytest.skip("io_stream does not support padding='same' for pools at the moment")

    run_model_test(model, cover_factor, data, io_type, backend, dir, aggressive)


if __name__ == '__main__':
    test_syn_players('PFlatten()', 10, 'floor', 'io_parallel', 0.5, True, 'vivado', 42)



import abc
from collections.abc import Callable
from warnings import warn

import numpy as np
import tensorflow as tf
from keras import backend as K
from tensorflow import keras

# Nice figure (Figure. 2 and 3) from https://www.researchgate.net/publication/226964494_Formalization_of_Fixed-Point_Arithmetic_in_HOL to illustrate the rounding and saturation modes.


def TRN(x):
    # Truncate towards negative infinity. Fast. Preferred when possible.
    return tf.floor(x)


def RND(x):
    # Round to nearest, ties to even.
    # Can be reduced to TRN with a bias.
    return tf.floor(x + 0.5)  # type:ignore


def RND_CONV(x):
    # towards nearest integer, ties to even.
    return tf.round(x)


def TRN_ZERO(x):
    # Truncate towards zero.
    sign = K.sign(x)
    return tf.floor(K.abs(x)) * sign


def RND_ZERO(x):
    # Round to nearest, ties to zero.
    sign = K.sign(x)
    return -tf.floor(-K.abs(x) + 0.5) * sign


def RND_MIN_INF(x):
    # Round to nearest, ties to negative infinity.
    return -tf.floor(-x + 0.5)  # type:ignore


def RND_INF(x):
    # Round to nearest, ties away from zero.
    sign = K.sign(x)
    return tf.floor(K.abs(x) + 0.5) * sign


def SAT(x, k, b):
    # Saturate between highest and lowest representable values.
    # Slow, but less trash results when overflows.
    high = 2**(b - k) - 1
    low = -(high + 1) * k
    return tf.clip_by_value(x, low, high)


def SAT_ZERO(x, k, b):
    # Overflow to zero.
    # Slow, and trash results when overflows. Why would anyone use this?
    high = 2**(b - k) - 1
    low = (-high - 1) * k
    mask = tf.cast((x <= high) & (x >= low), 'float32')
    return x * mask


def SAT_SYM(x, k, b):
    # Saturate between highest and lowest representable values when unsigned; between highest and -highest when signed.
    # Slow, but less trash results when overflows. Smaller representable range than SAT.
    high = 2**(b - k) - 1
    low = -high * k
    return tf.clip_by_value(x, low, high)


def WRAP(x, k, b):
    # Wrap around.
    # Faaaast, but trash results when overflows. YOLO.
    high = 2**(b - k) - 1
    low = -(high + 1) * k
    return tf.math.floormod(x - low, high - low + 1) + low


def WRAP_SM(x, k, b):
    # High and low bounds are reflective. When overflows, can be less trash than WARP but still more trash than SAT.
    dtype = x.dtype
    high = 2**(b - k) - 1
    low = -(high + 1) * k
    interval = (high - low + 1) * 2
    mapped = K.cast(tf.math.floormod(x - high - 1, interval), 'float32')
    return K.cast(K.abs(mapped - interval / 2 + 0.5) - 0.5 + low, dtype)


RND_MAP = {
    'RND': RND,
    'RND_ZERO': RND_ZERO,
    'RND_MIN_INF': RND_MIN_INF,
    'RND_INF': RND_INF,
    'RND_CONV': RND_CONV,
    'TRN_ZERO': TRN_ZERO,
    'TRN': TRN,
}

SAT_MAP = {
    'SAT': SAT,
    'SAT_ZERO': SAT_ZERO,
    'SAT_SYM': SAT_SYM,
    'WRAP': WRAP,
    'WRAP_SM': WRAP_SM,
}


@tf.function(autograph=False, jit_compile=True)
def gfixed_quantizer(x, keep_negative, bits, integer_bits, RND='TRN', SAT='WRAP'):
    '''Generalized fixed point quantizer, should have the same behavior to ap_fixed/ap_ufixed. Support high granularity quantization and broadcasting of bitwidths. RND and SAT mode must be strings.'''

    keep_negative = tf.cast(keep_negative, 'float32')
    bits = tf.cast(bits, 'float32')
    integer_bits = tf.cast(integer_bits, dtype='float32')

    two = tf.constant(2, dtype='float32')
    float_bits = bits - integer_bits  # type:ignore
    scale = tf.pow(two, float_bits)

    scaled_input = x * scale
    rnd, sat = RND_MAP[RND], SAT_MAP[SAT]
    quantized = sat(rnd(scaled_input), keep_negative, bits)
    return quantized / scale * tf.cast(bits != 0, 'float32')


def gfixed(keep_negative, bits, integer_bits, RND='TRN', SAT='WRAP') -> Callable:
    def compute(x):
        return gfixed_quantizer(x, keep_negative, bits, integer_bits, RND, SAT)  # type:ignore
    return compute


def ufixed(bits, integer_bits, RND='TRN', SAT='WRAP') -> Callable:
    return gfixed(0, bits, integer_bits, RND, SAT)


def fixed(bits, integer_bits, RND='TRN', SAT='WRAP') -> Callable:
    return gfixed(1, bits, integer_bits, RND, SAT)


@keras.utils.register_keras_serializable(package='HGQ')
class FixedPointQuantizer(keras.layers.Layer, metaclass=abc.ABCMeta):

    def __init__(self, keep_negative, bits, integers, RND: str = 'TRN', SAT: str = 'WRAP', overrides: dict | None = None, **kwargs):

        zeros = bits == 0

        if not hasattr(keep_negative, 'shape'):
            keep_negative = np.int8(keep_negative)
            keep_negative = tf.constant([keep_negative], dtype='int8')
        if not hasattr(bits, 'shape'):
            bits = np.int8(bits)
            bits = tf.constant([bits], dtype='int8')
        if not hasattr(integers, 'shape'):
            integers = np.int8(integers)
            integers = tf.constant([integers], dtype='int8')

        keep_negative = tf.where(zeros, tf.zeros_like(keep_negative), keep_negative)
        integers = tf.where(zeros, tf.zeros_like(integers), integers)
        self.keep_negative = tf.Variable(keep_negative, dtype='int8', name='keep_negative', trainable=False)
        self.bits = tf.Variable(bits, dtype='int8', name='bits', trainable=False)
        self.integers = tf.Variable(integers, dtype='int8', name='integers', trainable=False)
        assert self.keep_negative.shape == self.bits.shape == self.integers.shape, f'Shapes mismatch: keep_negative, bits, and integers must have the same shape. Got {self.keep_negative.shape}, {self.bits.shape}, {self.integers.shape}.'

        self.RND = RND
        self.SAT = SAT
        self.overrides = overrides or {'layers': {}}
        kwargs.pop('trainable', None)
        self._quantizer_created = False

        super().__init__(trainable=False, **kwargs)

    def call(self, x, training=None):  # type:ignore
        assert not training, "Proxy model shall can not be trained!"
        if not self.built:
            self.build(x.shape)
        return gfixed_quantizer(x, self.keep_negative, self.bits, self.integers, self.RND, self.SAT)  # type:ignore

    @property
    def result_t_kif(self):
        k, i, f = self.keep_negative, self.integers - self.keep_negative, self.bits - self.integers  # type:ignore
        k, i, f = np.max(k), np.max(i), np.max(f)  # type:ignore
        return k, i, f

    @property
    def fusible(self):
        """Delete this quantizer if no heterogeneity is detected."""
        assert len(self._inbound_nodes) == 1, 'FixedPointQuantizer must not be reused. Create proxy model only via proviced functions.'
        last_layer = self._inbound_nodes[0].inbound_layers
        assert not isinstance(last_layer, list), f'FixedPointQuantizer has exactly one inbound layer. Got a list of {len(last_layer)} layers.'
        if len(last_layer._outbound_nodes) != 1:
            return False
        return not self.heterogeneous

    @property
    def heterogeneous(self):
        k0, b0, i0 = tf.reduce_max(self.keep_negative), tf.reduce_max(self.bits), tf.reduce_max(self.integers)
        if not tf.reduce_all(self.keep_negative == k0):
            return True
        if not tf.reduce_all(self.bits == b0):
            return True
        if not tf.reduce_all(self.integers == i0):
            return True
        return False

    def get_config(self):
        assert tf.reduce_all((self.keep_negative == 0) | (self.keep_negative == 1)), 'Illegal bitwidth config: keep_negative must be 0 or 1.'
        if not tf.reduce_all(self.bits >= 0):  # type:ignore
            warn('Illegal bitwidth config: bits must be non-negative.')
            self.bits.assign(tf.maximum(self.bits, 0))
        conf = super().get_config()
        conf['RND'] = self.RND
        conf['SAT'] = self.SAT
        conf['shape'] = tuple(self.bits.shape)
        overrides = self.overrides

        conf['overrides'] = overrides
        conf['fusible'] = self.fusible
        return conf

    @classmethod
    def from_config(cls, config: dict):
        dummy_v = np.full(config.pop('shape'), -128, dtype='int8')
        keep_negative = K.variable(dummy_v, dtype='int8', name='keep_negative')
        bits = K.variable(dummy_v, dtype='int8', name='bits')
        integers = K.variable(dummy_v, dtype='int8', name='integers')
        config.pop('fusible', None)
        return cls(keep_negative, bits, integers, **config)



from .convert import to_proxy_model
from .fixed_point_quantizer import FixedPointQuantizer, fixed, gfixed, gfixed_quantizer, ufixed
# Register plugins
from .plugins import init_all
from .unary_lut import UnaryLUT

init_all()



import sys

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

from . import _patch


class Shutup:
    def write(self, s):
        pass

    def flush(self):
        pass

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = self
        self._original_stderr = sys.stderr
        sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr

    def close(self):
        pass


shutup = Shutup()

with shutup:

    from .bops import CalibratedBOPs, FreeBOPs, ResetMinMax, trace_minmax
    from .layers import HConv1D, HConv1DBatchNorm, HConv2D, HConv2DBatchNorm, HDense, HDenseBatchNorm, HQuantize, PAvgPool1D, PAvgPool2D, PConcatenate, PFlatten, PMaxPool1D, PMaxPool2D, PReshape, Signature
    from .proxy import to_proxy_model
    from .quantizer import HGQ
    from .utils import get_default_kq_conf, get_default_paq_conf, set_default_kq_conf, set_default_paq_conf



import keras
import numpy as np
import pytest
from helpers import get_test_dir, run_model_test, set_seed

from HGQ import FreeBOPs, HDense, HDenseBatchNorm, HQuantize, ResetMinMax, get_default_kq_conf, get_default_paq_conf, set_default_kq_conf, set_default_paq_conf, trace_minmax


def get_model(io_type: str, rnd_strategy) -> keras.models.Model:
    beta = 5e-5

    paq_conf = get_default_paq_conf()

    paq_conf['skip_dims'] = 'all' if io_type == 'io_stream' else 'batch'
    paq_conf['rnd_strategy'] = rnd_strategy
    paq_conf['init_bw'] = 8

    set_default_paq_conf(paq_conf)

    kq_conf = get_default_kq_conf()

    kq_conf['init_bw'] = 8

    set_default_kq_conf(kq_conf)

    model = keras.models.Sequential([
        HQuantize(input_shape=(1,)),
        HDense(16, activation='relu', beta=beta),
        HDense(16, activation='relu', beta=beta),
        HDense(16, activation='relu', beta=beta),
        HDense(1),
    ])

    loss = keras.losses.MSE
    opt = keras.optimizers.Adam(learning_rate=0.01)

    model.compile(loss=loss, optimizer=opt)
    return model


def data():
    x = np.linspace(-np.pi, np.pi, 10000).astype(np.float32).reshape(-1, 1)
    y = np.sin(x)
    return x, y


@pytest.mark.parametrize('io_type', ['io_stream', 'io_parallel'])
@pytest.mark.parametrize('seed', [42, 114514, 1919810])
@pytest.mark.parametrize('rnd_strategy', ['standard_round', 'stochastic_round', 'fast_uniform_noise_injection', 'floor'])
def test_end2end(io_type: str, seed: int, rnd_strategy: str):

    set_seed(seed)

    model = get_model(io_type, rnd_strategy)
    x, y = data()

    callbacks = [FreeBOPs(), ResetMinMax()]
    logs = model.fit(x, y, epochs=10, batch_size=500, callbacks=callbacks, verbose=0)  # type: ignore

    mse = np.mean((model(x) - y)**2)
    assert mse < 0.01 if 'io_stream' else 0.005, f'Obtained MSE={mse}, expected < {0.01 if "io_stream" else 0.005}'

    bops: float = trace_minmax(model, x, verbose=False)  # type: ignore
    assert bops < 25000 if 'io_stream' else 20000, f'Obtained BOPs={bops}, expected < {25000 if "io_stream" else 20000}'
    assert np.mean(np.sign(np.diff(logs.history['bops']))) < -0.5, 'BOPs should decrease during training, but it does not'

    run_model_test(model, 1.0, x, io_type, 'vivado', get_test_dir(), True)



import numpy as np
import tensorflow as tf
from keras.saving import register_keras_serializable

from .base import HLayerBase
from .batchnorm_base import HBatchNormBase


@register_keras_serializable(package="HGQ")
class HDense(HLayerBase, tf.keras.layers.Dense):
    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor=None,
        **kwargs,
    ):
        super().__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            **kwargs
        )
        self.parallel_factor = parallel_factor
        """Acting like `parallelization_factor` in conv, but for multi-dimension dense layer (matmul operation)."""

    def forward(self, x, training=None, record_minmax=None):

        a, kq = self.jit_forward(x, training=training, record_minmax=record_minmax)  # type: ignore

        input_bw = self.input_bw
        if input_bw is not None:
            kernel_bw = self._kernel_bw(kq)  # type: ignore
            bops = tf.reduce_sum(tf.matmul(input_bw, kernel_bw))
            self.bops.assign(bops)
            bops = tf.cast(bops, tf.float32) * self.beta  # type: ignore
            self.add_loss(tf.convert_to_tensor(bops))
        return a

    @tf.function(jit_compile=True)
    def jit_forward(self, x, training=None, record_minmax=None):

        kq = self.kq(self.fused_kernel, training, False)  # type: ignore
        z = tf.matmul(x, kq)
        if self.use_bias:
            b = self.paq.bias_forward(self.fused_bias, training, self.channel_loc)  # type: ignore
            z = tf.nn.bias_add(z, b)
        z = self.paq(z, training, record_minmax)  # type: ignore
        a = self.activation(z)  # type: ignore

        return a, kq

    @property
    def compute_exact_bops(self):
        kernel_bw = self.kernel_bw_exact
        input_bw = self.input_bw_exact
        bops = np.sum(np.matmul(input_bw, kernel_bw))  # type: ignore
        self.bops.assign(tf.constant(bops, dtype=tf.float32))
        return bops

    def post_build(self, input_shape):
        super().post_build(input_shape)
        if self.parallel_factor is None:
            if len(input_shape) > 2:
                self.parallel_factor = int(np.prod(input_shape[1:-1]))


@register_keras_serializable(package="HGQ")
class HDenseBatchNorm(HDense, HBatchNormBase):

    def __init__(
        self,
        units,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(
            units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            axis=-1,
            momentum=momentum,
            epsilon=epsilon,
            center=center,
            scale=scale,
            beta_initializer=beta_initializer,
            gamma_initializer=gamma_initializer,
            moving_mean_initializer=moving_mean_initializer,
            moving_variance_initializer=moving_variance_initializer,
            beta_regularizer=beta_regularizer,
            gamma_regularizer=gamma_regularizer,
            beta_constraint=beta_constraint,
            gamma_constraint=gamma_constraint,
            **kwargs,
        )
        self._delayed_kernel_bits_adaption = True

    def forward(self, x, training=None, record_minmax=None):

        if training and self._do_adapt_kernel_bits and self._delayed_kernel_bits_adaption:
            self.adapt_fused_bn_kernel_bw_bits(x)
            self._do_adapt_kernel_bits = False

        if training:
            a, kq = self.bn_train_jit_forward(x, True, record_minmax=record_minmax)  # type: ignore
        else:
            a, kq = self.jit_forward(x, False, record_minmax=record_minmax)  # type: ignore

        input_bw = self.input_bw
        if input_bw is not None:
            kernel_bw = self._kernel_bw(kq)  # type: ignore
            bops = tf.reduce_sum(tf.matmul(input_bw, kernel_bw))
            self.bops.assign(bops)
            bops = tf.cast(bops, tf.float32) * self.beta  # type: ignore
            self.add_loss(tf.convert_to_tensor(bops))
        return a

    @tf.function(jit_compile=True)
    def bn_train_jit_forward(self, x, training, record_minmax=None):

        if self.scale:
            kq = self.kernel
        else:
            kq = self.kq(self.kernel, training, False)  # type: ignore

        z = tf.matmul(x, kq)

        if self.center:
            mean = tf.math.reduce_mean(z, axis=self._reduction_axis) + self.bias
            self.moving_mean.assign(self.momentum * self.moving_mean + (1 - self.momentum) * mean)

        if self.scale:
            var = tf.math.reduce_variance(z, axis=self._reduction_axis)
            self.moving_variance.assign(self.momentum * self.moving_variance + (1 - self.momentum) * var)
            scale = self.bn_gamma * tf.math.rsqrt(var + self.epsilon)
            train_fused_kernel = self.kernel * scale
            kq = self.kq(train_fused_kernel, training, False)  # type: ignore
            z = tf.matmul(x, kq)
        else:
            scale = tf.constant(1.0, dtype=self.dtype)

        if self.center:
            train_fused_bias = self.bias - mean * scale  # type: ignore
            bq = self.paq.bias_forward(train_fused_bias, training, self.channel_loc)  # type: ignore
            z = tf.nn.bias_add(z, bq)  # type: ignore

        z = self.paq(z, training, record_minmax)  # type: ignore
        a = self.activation(z)  # type: ignore

        return a, kq



import numpy as np
import pytest
import tensorflow as tf
from helpers import get_test_dir, run_model_test, set_seed
from tensorflow import keras

import HGQ
from HGQ import get_default_paq_conf, set_default_paq_conf
from HGQ.layers import HQuantize, PReshape


def create_model(layer: str, rnd_strategy: str, io_type: str):
    pa_config = get_default_paq_conf()
    pa_config['rnd_strategy'] = rnd_strategy
    pa_config['skip_dims'] = 'all' if io_type == 'io_stream' else 'batch'
    set_default_paq_conf(pa_config)

    inp = keras.Input(shape=(16))
    if 'Add' in layer:
        _inp = [HQuantize()(inp)] * 2
    elif 'Conv2D' in layer:
        _inp = PReshape((4, 4, 1))(HQuantize()(inp))
    elif 'Conv1D' in layer:
        _inp = PReshape((16, 1))(HQuantize()(inp))
    elif 'Dense' in layer or 'Activation' in layer:
        _inp = HQuantize()(inp)
    else:
        raise Exception(f'Please add test for {layer}')

    _layer = eval('HGQ.layers.' + layer)
    if hasattr(_layer, 'bias') and _layer.bias is not None:
        bias: tf.Variable = _layer.bias
        bias.assign(tf.constant(np.random.uniform(-4, 4, _layer.bias.shape).astype(np.float32)))

    out = _layer(_inp)
    model = keras.Model(inp, out)

    for layer in model.layers:
        # Randomize weight bitwidths
        if hasattr(layer, 'kq'):
            fbw: tf.Variable = layer.kq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 8, fbw.shape).astype(np.float32)))
        # And activation bitwidths
        if hasattr(layer, 'paq'):
            fbw: tf.Variable = layer.paq.fbw
            fbw.assign(tf.constant(np.random.uniform(2, 8, fbw.shape).astype(np.float32)))

    return model


def get_data(N: int, sigma: float, max_scale: float, seed):
    rng = np.random.default_rng(seed)
    a1 = rng.normal(0, sigma, (N, 16)).astype(np.float32)
    a2 = rng.uniform(0, max_scale, (1, 16)).astype(np.float32)
    return (a1 * a2).astype(np.float32)


def softmax_cond(proxy, hls):
    match_precent = np.mean(np.argmax(proxy, axis=1) == np.argmax(hls, axis=1))
    assert match_precent > 0.90, f"Proxy-HLS4ML mismatch: {(1-match_precent) * 100}% of samples are different. Sample: {proxy[:5]} vs {hls[:5]}"


def custom_fn(x):
    return tf.sin(x)**2. - x  # type: ignore


@pytest.mark.parametrize('layer',
                         ["HDense(10)",
                          "HDense(10, use_bias=False)",
                          "HDenseBatchNorm(10)",
                          "HConv1D(2, 3, padding='same')",
                          "HConv1D(2, 3, padding='valid')",
                          "HConv1D(2, 3, padding='valid', use_bias=False)",
                          "HConv1D(2, 3, padding='valid', strides=2)",
                          "HConv1D(2, 3, padding='same', strides=2)",
                          "HConv1DBatchNorm(2, 3, padding='valid')",
                          "HConv2D(2, (3,3), padding='same')",
                          "HConv2D(2, (3,3), padding='valid')",
                          "HConv2D(2, (3,3), padding='valid', use_bias=False)",
                          "HConv2D(2, (3,3), padding='valid', strides=2)",
                          "HConv2D(2, (3,3), padding='same', strides=2)",
                          "HConv2DBatchNorm(2, (3,3), padding='valid')",
                          "HAdd()",
                          "HActivation('relu')",
                          #   "HActivation('leaky_relu')",
                          "HActivation('tanh')",
                          "HActivation('sigmoid')",
                          "HActivation('softmax')",
                          "HActivation(custom_fn)",
                          ]
                         )
@pytest.mark.parametrize("N", [1000])
@pytest.mark.parametrize("rnd_strategy", ['auto', 'standard_round', 'floor'])
@pytest.mark.parametrize("io_type", ['io_parallel', 'io_stream'])
@pytest.mark.parametrize("cover_factor", [0.25, 1.0])
@pytest.mark.parametrize("aggressive", [True, False])
@pytest.mark.parametrize("backend", ['vivado', 'vitis'])
@pytest.mark.parametrize("seed", [42])
def test_syn_hlayers(layer, N: int, rnd_strategy: str, io_type: str, cover_factor: float, aggressive: bool, backend: str, seed: int):
    dir = get_test_dir()
    set_seed(seed)
    model = create_model(layer=layer, rnd_strategy=rnd_strategy, io_type=io_type)
    data = get_data(N, 1, 1, seed)

    test_grad = N > 100
    cond = None
    if 'softmax' in layer:
        cond = softmax_cond
        cover_factor = max(cover_factor, 0.5)  # Softmax table size is sensitive to range

    skip_sl_test = 'custom' in layer
    run_model_test(model,
                   cover_factor, data,
                   io_type,
                   backend,
                   dir,
                   aggressive,
                   test_gard=test_grad,
                   cond=cond,
                   skip_sl_test=skip_sl_test,
                   )


if __name__ == '__main__':
    test_syn_hlayers('HActivation("tanh")', 10, 'floor', 'io_parallel', 0.5, True, 'vivado', 42)



from functools import partialmethod

import tensorflow as tf
from keras.saving import register_keras_serializable
from keras.src.layers.convolutional.base_conv import Conv

from ..utils import warn
from .base import HLayerBase
from .batchnorm_base import HBatchNormBase


class HConv(HLayerBase, Conv):
    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        conv_op=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        **kwargs,
    ):
        """parallel_factor: number of parallel kernel operation to be used. Only used in bops estimation."""
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            conv_op=conv_op,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            **kwargs,
        )
        if conv_op is not None:
            warn(r'Warning: conv_op is defined. This will override the default convolution operation.')
            self.convolution_op = conv_op
        self.parallel_factor = tf.constant(parallel_factor, dtype=tf.float32)
        self.channel_loc = 1 if self.data_format == "channels_first" else -1

    def post_build(self, input_shape):
        r = super().post_build(input_shape)
        self.strides = list(self.strides)
        self.dilation_rate = list(self.dilation_rate)
        output_shape = self.compute_output_shape(input_shape)
        self.total_channels = tf.cast(tf.reduce_prod(output_shape[1:-1]), dtype=tf.float32)
        self.kq.degeneracy *= float(self.parallel_factor)
        return r

    @tf.function(jit_compile=True)
    def convolution_op(self, inputs, kernel):
        if self.padding == "causal":
            tf_padding = "VALID"  # Causal padding handled in `call`.
        elif isinstance(self.padding, str):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding

        return tf.nn.convolution(
            inputs,
            kernel,
            strides=self.strides,
            padding=tf_padding,
            dilations=self.dilation_rate,
            data_format=self._tf_data_format,
            name=self.__class__.__name__,
        )

    def forward(self, x, training=None, record_minmax=None):

        a, kq = self.jit_forward(x, training=training, record_minmax=record_minmax)  # type: ignore

        input_bw = self.input_bw
        if input_bw is not None:
            kernel_bw = self._kernel_bw(kq)  # type: ignore
            bops = tf.reduce_sum(self.convolution_op(self.input_bw, kernel_bw))  # type: ignore
            bops = bops * self.parallel_factor / self.total_channels
            self.bops.assign(bops)
            bops = tf.cast(bops, tf.float32) * self.beta
            self.add_loss(tf.convert_to_tensor(bops))
        return a

    @tf.function(jit_compile=True)
    def jit_forward(self, x, training=None, record_minmax=None):

        kq = self.kq(self.fused_kernel, training, False)  # type: ignore
        z = self.convolution_op(x, kq)  # type: ignore
        if self.use_bias:
            b = self.paq.bias_forward(self.fused_bias, training)  # type: ignore
            z = tf.nn.bias_add(z, b, data_format=self._tf_data_format)
        z = self.paq(z, training, record_minmax)  # type: ignore
        a = self.activation(z)  # type: ignore

        return a, kq

    @property
    def compute_exact_bops(self):
        kernel_bw = tf.constant(self.kernel_bw_exact, dtype=tf.float32)
        input_bw = self.input_bw_exact
        bops = int(tf.reduce_sum(self.convolution_op(input_bw, kernel_bw)).numpy()) * int(self.parallel_factor.numpy()) / int(self.total_channels.numpy())  # type: ignore
        self.bops.assign(tf.constant(bops, dtype=tf.float32))
        return bops


@register_keras_serializable(package="HGQ")
class HConv2D(HConv):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        **kwargs,
    ):
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            parallel_factor=parallel_factor,
            **kwargs,
        )


@register_keras_serializable(package="HGQ")
class HConv1D(HConv):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        **kwargs,
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            parallel_factor=parallel_factor,
            **kwargs,
        )


class HConvBatchNorm(HConv, HBatchNormBase):
    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        conv_op=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        channel_loc = 1 if data_format == "channels_first" else -1
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            conv_op=conv_op,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            parallel_factor=parallel_factor,
            axis=channel_loc,
            momentum=momentum,
            epsilon=epsilon,
            center=center,
            scale=scale,
            beta_initializer=beta_initializer,
            gamma_initializer=gamma_initializer,
            moving_mean_initializer=moving_mean_initializer,
            moving_variance_initializer=moving_variance_initializer,
            beta_regularizer=beta_regularizer,
            gamma_regularizer=gamma_regularizer,
            beta_constraint=beta_constraint,
            gamma_constraint=gamma_constraint,
            **kwargs,
        )

    def forward(self, x, training=None, record_minmax=None):

        if training and self._do_adapt_kernel_bits and self._delayed_kernel_bits_adaption:
            self.adapt_fused_bn_kernel_bw_bits(x)
            self._do_adapt_kernel_bits = False

        if training:
            a, kq = self.bn_train_jit_forward(x, True, record_minmax=record_minmax)  # type: ignore
        else:
            a, kq = self.jit_forward(x, False, record_minmax=record_minmax)  # type: ignore

        input_bw = self.input_bw
        if input_bw is not None:
            kernel_bw = self._kernel_bw(kq)  # type: ignore
            bops = tf.reduce_sum(self.convolution_op(self.input_bw, kernel_bw))  # type: ignore
            bops = bops * self.parallel_factor / self.total_channels
            self.bops.assign(bops)
            bops = tf.cast(bops, tf.float32) * self.beta
            self.add_loss(tf.convert_to_tensor(bops))
        return a

    @tf.function(jit_compile=True)
    def bn_train_jit_forward(self, x, training, record_minmax=None):

        if self.scale:
            kq = self.kernel
        else:
            kq = self.kq(self.kernel, training, False)  # type: ignore

        z = self.convolution_op(x, kq)  # type: ignore

        if self.center:
            mean = tf.math.reduce_mean(z, axis=self._reduction_axis) + self.bias
            self.moving_mean.assign(self.momentum * self.moving_mean + (1 - self.momentum) * mean)

        if self.scale:
            var = tf.math.reduce_variance(z, axis=self._reduction_axis)
            self.moving_variance.assign(self.momentum * self.moving_variance + (1 - self.momentum) * var)
            scale = self.bn_gamma * tf.math.rsqrt(var + self.epsilon)
            train_fused_kernel = self.kernel * scale
            kq = self.kq(train_fused_kernel, training, False)  # type: ignore
            z = self.convolution_op(x, kq)  # type: ignore
        else:
            scale = tf.constant(1.0, dtype=self.dtype)

        if self.center:
            train_fused_bias = self.bias - mean * scale  # type: ignore
            bq = self.paq.bias_forward(train_fused_bias, training, self.channel_loc)  # type: ignore
            z = tf.nn.bias_add(z, bq, data_format=self._tf_data_format)  # type: ignore

        z = self.paq(z, training, record_minmax)  # type: ignore
        a = self.activation(z)  # type: ignore

        return a, kq


@register_keras_serializable(package="HGQ")
class HConv2DBatchNorm(HConvBatchNorm):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        conv_op=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(
            rank=2,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            conv_op=conv_op,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            parallel_factor=parallel_factor,
            momentum=momentum,
            epsilon=epsilon,
            center=center,
            scale=scale,
            beta_initializer=beta_initializer,
            gamma_initializer=gamma_initializer,
            moving_mean_initializer=moving_mean_initializer,
            moving_variance_initializer=moving_variance_initializer,
            beta_regularizer=beta_regularizer,
            gamma_regularizer=gamma_regularizer,
            beta_constraint=beta_constraint,
            gamma_constraint=gamma_constraint,
            **kwargs,
        )


@register_keras_serializable(package="HGQ")
class HConv1DBatchNorm(HConvBatchNorm):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        conv_op=None,
        kq_conf=None,
        paq_conf=None,
        beta=0.,
        parallel_factor: int = 1,
        momentum=0.99,
        epsilon=1e-3,
        center=True,
        scale=True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        moving_mean_initializer="zeros",
        moving_variance_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(
            rank=1,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            trainable=trainable,
            name=name,
            conv_op=conv_op,
            kq_conf=kq_conf,
            paq_conf=paq_conf,
            beta=beta,
            parallel_factor=parallel_factor,
            momentum=momentum,
            epsilon=epsilon,
            center=center,
            scale=scale,
            beta_initializer=beta_initializer,
            gamma_initializer=gamma_initializer,
            moving_mean_initializer=moving_mean_initializer,
            moving_variance_initializer=moving_variance_initializer,
            beta_regularizer=beta_regularizer,
            gamma_regularizer=gamma_regularizer,
            beta_constraint=beta_constraint,
            gamma_constraint=gamma_constraint,
            **kwargs,
        )
